"""This tool enables easy generation of random features for
approximate kernel k-means clustering or for other tasks."""
import numpy as np
try:
    import cupy as cp
except:
    print("CuPy not detected. xGPR will run in CPU-only mode.")

from .constants import constants
from .auxiliary_baseclass import AuxiliaryBaseclass



class KernelFGen(AuxiliaryBaseclass):
    """A tool for generating random features using
    a selected kernel, typically for use in e.g.
    kernel k-means clustering."""

    def __init__(self, num_rffs:int, hyperparams, num_features:int,
            kernel_choice:str = "RBF", device:str = "cpu",
            kernel_settings:dict = constants.DEFAULT_KERNEL_SPEC_PARMS,
            random_seed:int = 123, verbose:bool = True, num_threads:int = 2):
        """The constructor.

        Args:
            num_rffs (int): The number of random Fourier features
                to use for the auxiliary device.
            hyperparams (np.ndarray): A numpy array containing the kernel-specific
                hyperparameter. If you have fitted an xGPR model, the first
                hyperparameter is in general not kernel specific, so
                my_model.get_hyperparams()[1:] will retrieve the hyperparameters you
                need. For most kernels there is only one kernel-specific hyperparameter.
                For kernels with no kernel-specific hyperparameter (e.g. polynomial
                kernels), this argument is ignored.
            num_features (int): The number of features in your input data. This
                should be the last dimension of a typical input array.
            kernel_choice (str): The kernel that the model will use.
                Must be in kernels.kernel_list.KERNEL_NAME_TO_CLASS.
            device (str): Determines whether calculations are performed on
                'cpu' or 'cuda'. The initial entry can be changed later
                (i.e. model can be transferred to a different device).
                Defaults to 'cpu'.
            kernel_settings (dict): Contains kernel-specific parameters --
                e.g. 'matern_nu' for the nu for the Matern kernel, or 'conv_width'
                for sequence / timeseries kernels.
            random_seed (int): A seed for the random number generator.
            verbose (bool): If True, regular updates are printed
                during fitting and tuning. Defaults to True.
            num_threads (int): The number of threads to use for random feature generation
                if running on CPU. If running on GPU, this argument is ignored.
        """
        super().__init__(num_rffs, hyperparams, num_features,
                        kernel_choice, device, kernel_settings,
                        random_seed, verbose, num_threads)


    def predict(self, input_x, sequence_lengths = None, chunk_size:int = 2000):
        """Generates random features for the input.

        Args:
            input_x (np.ndarray): A numpy array of xvalues for which RFFs will
                be generated.
            sequence_lengths: None if you are using a fixed-vector kernel (e.g.
                RBF) and a 1d array of the number of elements in each sequence /
                nodes in each graph if you are using a graph or Conv1d kernel.
            chunk_size (int): How many datapoints to process at a time. The
                results for all datapoints are returned simultaneously regardless --
                this parameter just affects memory consumption (smaller chunk_size
                = less memory) and speed (larger chunk_size = slightly faster).
        Returns:
            preds (np.ndarray): A 2d array of RFFs where the second dimension is
                the number of rffs.

        Raises:
            ValueError: A ValueError is raised if inappropriate inputs are
                supplied.
        """
        self.pre_prediction_checks(input_x, sequence_lengths)
        preds = []
        for i in range(0, input_x.shape[0], chunk_size):
            cutoff = min(i + chunk_size, input_x.shape[0])
            if sequence_lengths is not None:
                preds.append(self.kernel.transform_x(input_x[i:cutoff, ...],
                    sequence_lengths[i:cutoff]))
            else:
                preds.append(self.kernel.transform_x(input_x[i:cutoff, ...]))

        if self.device == "cuda":
            preds = cp.asnumpy(cp.vstack(preds))
        else:
            preds = np.vstack(preds)
        return preds



"""Describes the SORFKernelBaseclass, which is used by the main
fixed-vector kernels -- RBF, Matern.

All of these share some methods and attributes in common
which are stored here to avoid redundancy."""
from abc import ABC
from math import ceil

import numpy as np
from scipy.stats import chi

from xGPR.xgpr_cpu_rfgen_cpp_ext import cpuRBFFeatureGen, cpuRBFGrad
try:
    import cupy as cp
    from xGPR.xgpr_cuda_rfgen_cpp_ext import cudaRBFFeatureGen, cudaRBFGrad
except:
    pass

from ..kernel_baseclass import KernelBaseclass



class SORFKernelBaseclass(KernelBaseclass, ABC):
    """The baseclass for structured orthogonal random features (SORF)
    kernels that accept fixed-vector inputs. Since it inherits from
    KernelBaseclass, it includes the attributes of that class. Only
    additional attributes unique to this class are described here.

    Attributes:
        radem_diag: The diagonal matrices for the SORF transform. Type is int8.
        chi_arr: A diagonal array whose elements are drawn from the chi
            distribution. Ensures the marginals of the matrix resulting
            from S H D1 H D2 H D3 are correct.
    """

    def __init__(self, num_rffs, xdim, num_threads = 2,
                            sine_cosine_kernel = True, random_seed = 123,
                            double_precision = False,
                            kernel_spec_parms = {}):
        """Constructor for the SORFKernelBaseclass. Calls the KernelBaseclass
        constructor first.

        Args:
            num_rffs (int): The user-requested number of random Fourier features.
                For sine-cosine kernels (RBF, Matern), this will be saved by the
                class as num_rffs.
            xdim (tuple): The dimensions of the input. Either (N, D) or (N, M, D)
                where N is the number of datapoints, D is number of features
                and M is number of timepoints or sequence elements (convolution
                kernels only).
            num_threads (int): The number of threads to use for generating random
                features if running on CPU. If running on GPU, this is ignored.
            random_seed (int): The seed to the random number generator.
            sine_cosine_kernel (bool): If True, the kernel is a sine-cosine kernel,
                meaning it will sample self.num_freqs frequencies and use the sine
                and cosine of each to generate twice as many features
                (self.num_rffs). sine-cosine kernels only accept num_rffs
                that are even numbers.
            double_precision (bool): If True, generate random features in double precision.
                Otherwise, generate as single precision.
            kernel_spec_parms (dict): A dictionary of other kernel-specific settings.

        Raises:
            ValueError: If a non 2d input array dimensionality is supplied.
        """
        super().__init__(num_rffs, xdim, num_threads = num_threads,
                sine_cosine_kernel = sine_cosine_kernel,
                double_precision = double_precision,
                kernel_spec_parms = kernel_spec_parms)
        if len(xdim) != 2:
            raise ValueError("The dimensionality of the input is inappropriate for "
                        "the kernel you have selected.")

        padded_dims = 2**ceil(np.log2(max(xdim[-1], 2)))

        radem_array = np.asarray([-1,1], dtype=np.int8)
        rng = np.random.default_rng(random_seed)
        if padded_dims < self.num_freqs:
            nblocks = ceil(self.num_freqs / padded_dims)
        else:
            nblocks = 1
        self.radem_diag = rng.choice(radem_array, size=(3, 1,
                nblocks * padded_dims), replace=True)
        self.chi_arr = chi.rvs(df=padded_dims, size=self.num_freqs,
                            random_state = random_seed)
        if not self.double_precision:
            self.chi_arr = self.chi_arr.astype(np.float32)



    def kernel_specific_set_device(self, new_device):
        """Called by parent class when device is changed. Moves
        some of the object parameters to the appropriate device
        and updates self.cosfunc, self.sinfunc, which are
        convenience references to np.cos / np.sin or cp.cos
        / cp.sin."""
        if new_device == "cpu":
            if not isinstance(self.radem_diag, np.ndarray):
                self.radem_diag = cp.asnumpy(self.radem_diag)
                self.chi_arr = cp.asnumpy(self.chi_arr)
        else:
            self.radem_diag = cp.asarray(self.radem_diag)
            self.chi_arr = cp.asarray(self.chi_arr)



    def kernel_specific_transform(self, input_x, sequence_length = None):
        """Combines the two steps involved in random feature generation
        to generate random features.

        Args:
            xtrans: Either a cupy or numpy array containing the input.
            sequence_length: Accepted for consistency with baseclass and
                kernels that use this argument but is not used by this
                class of kernels and is therefore ignored.

        Returns:
            xtrans: A cupy or numpy array containing the generated features.
        """
        input_x *= self.hyperparams[1]
        if self.device == "cpu":
            output_x = np.zeros((input_x.shape[0], self.num_rffs), np.float64)
            cpuRBFFeatureGen(input_x, output_x, self.radem_diag, self.chi_arr,
                self.num_threads, self.fit_intercept)
        else:
            output_x = cp.zeros((input_x.shape[0], self.num_rffs), cp.float64)
            cudaRBFFeatureGen(input_x, output_x, self.radem_diag, self.chi_arr,
                self.fit_intercept)
        return output_x


    def kernel_specific_set_hyperparams(self):
        """Provided for consistency with baseclass. This
        kernel has no kernel-specific properties that must
        be reset after hyperparameters are changed."""
        return


    def kernel_specific_gradient(self, input_x, sequence_length = None):
        """The gradient for kernel-specific hyperparameters is calculated
        using an array (dz_dsigma) specific to each kernel.

        Args:
            input_x: A cupy or numpy array containing the raw input data.
            sequence_length: Accepted for consistency with baseclass and
                kernels that use this argument but is not used by this
                class of kernels and is therefore ignored.

        Returns:
            output_x: A cupy or numpy array containing the random feature
                representation of the input.
            dz_dsigma: A cupy or numpy array containing the derivative of
                output_x with respect to the kernel-specific hyperparameters.
        """
        if self.device == "cpu":
            output_x = np.zeros((input_x.shape[0], self.num_rffs), np.float64)
            dz_dsigma = np.zeros((input_x.shape[0], self.num_rffs, 1), np.float64)
            cpuRBFGrad(input_x, output_x, dz_dsigma, self.radem_diag, self.chi_arr,
                self.hyperparams[1], self.num_threads, self.fit_intercept)
        else:
            output_x = cp.zeros((input_x.shape[0], self.num_rffs), cp.float64)
            dz_dsigma = cp.zeros((input_x.shape[0], self.num_rffs, 1), cp.float64)
            cudaRBFGrad(input_x, output_x, dz_dsigma, self.radem_diag, self.chi_arr,
                self.hyperparams[1], self.fit_intercept)
        return output_x, dz_dsigma



"""Contains functions for building OnlineDataset and OfflineDataset objects.

When constructing a dataset, we should check to ensure the user has passed
valid data (to avoid cryptic errors during tuning or fitting). The functions
here provide the user with a way to construct online or offline (in memory
or on disk) datasets as appropriate while checking to ensure validity.
"""
import numpy as np

from .online_data_handling import OnlineDataset
from .offline_data_handling import OfflineDataset


def build_regression_dataset(xdata, ydata, sequence_lengths = None,
        chunk_size:int = 2000, normalize_y:bool = True):
    """Builds a dataset intended for use for regression.

    Args:
        xdata (np.ndarray): Either a numpy array containing the x-values
            or a list of valid filepaths to .npy numpy arrays containing
            the x-data. If the latter, these must all be either
            2d arrays where shape[1] is the number of features or
            3d arrays where shape[1] is the number of timepoints
            or sequence elements and shape[2] is the number of
            features for each timepoint / sequence element. They
            should be one or the other, not both.
        ydata (np.ndarray): Either a numpy array containing the y-values
            or a list of valid filepaths to .npy numpy
            arrays containing the y-data. These must all be 1d
            arrays. The number of datapoints in each should
            match the corresponding element of xlist.
        sequence_lengths (np.ndarray): Either (a) a numpy array containing
            the length of each sequence / timeseries / graph for each datapoint or
            (b) a list of valid filepaths to .npy files containing the
            length of each sequence / timeseries / graph or (c) None. If you are
            not using a sequence kernel, this must always be None. If you are
            using a sequence / graph kernel, this cannot be None. In other words,
            this is required if using a sequence / graph kernel and must be
            None otherwise.
        chunk_size (int): The maximum size of data chunks that
            will be returned to callers. Limits memory consumption.
            Defaults to 2000. If working with files on disk, the
            files will be checked to ensure that none of them is
            > than this size.
        normalize_y (bool): If True, y values are normalized. Generally a
            good idea, unless you have already selected hyperparameters based
            on prior knowledge.

    Returns:
        dataset: An object of class OnlineDataset or OfflineDataset
            that can be passed to the hyperparameter tuning
            and fitting routines of the model classes.

    Raises:
        ValueError: A ValueError is raised if inappropriate argument
            types are supplied.
    """
    if isinstance(xdata, list) and isinstance(ydata, list):
        return build_offline_np_dataset(xdata, ydata, sequence_lengths,
                chunk_size, normalize_y, task_type = "regression")
    if isinstance(xdata, np.ndarray) and isinstance(ydata, np.ndarray):
        return build_online_dataset(xdata, ydata, sequence_lengths,
                chunk_size, normalize_y, task_type = "regression")
    raise ValueError("Unexpected argument types to build_regression_dataset.")



def build_classification_dataset(xdata, ydata,
        sequence_lengths = None, chunk_size:int = 2000):
    """Builds a dataset intended for use for classification.

    Args:
        xdata (np.ndarray): Either a numpy array containing the x-values
            or a list of valid filepaths to .npy numpy arrays containing
            the x-data. If the latter, these must all be either
            2d arrays where shape[1] is the number of features or
            3d arrays where shape[1] is the number of timepoints
            or sequence elements and shape[2] is the number of
            features for each timepoint / sequence element. They
            should be one or the other, not both.
        ydata (np.ndarray): Either a numpy array containing the y-values
            or a list of valid filepaths to .npy numpy
            arrays containing the y-data. These must all be 1d
            arrays. The number of datapoints in each should
            match the corresponding element of xlist. The category
            for each datapoint should be specified as an integer in
            [0, max category]. E.g. if there are three categories
            they will be numbereed 0, 1, 2.
        sequence_lengths (np.ndarray): Either (a) a numpy array containing
            the length of each sequence / timeseries / graph for each datapoint or
            (b) a list of valid filepaths to .npy files containing the
            length of each sequence / timeseries / graph or (c) None. If you are
            not using a sequence kernel, this must always be None. If you are
            using a sequence / graph kernel, this cannot be None. In other words,
            this is required if using a sequence / graph kernel and must be
            None otherwise.
        chunk_size (int): The maximum size of data chunks that
            will be returned to callers. Limits memory consumption.
            Defaults to 2000. If working with files on disk, the
            files will be checked to ensure that none of them is
            > than this size.

    Returns:
        dataset: An object of class OnlineDataset or OfflineDataset
            that can be passed to the hyperparameter tuning
            and fitting routines of the model classes.

    Raises:
        ValueError: A ValueError is raised if inappropriate argument
            types are supplied.
    """
    if isinstance(xdata, list) and isinstance(ydata, list):
        return build_offline_np_dataset(xdata, ydata, sequence_lengths,
                chunk_size, normalize_y = False, task_type = "classification")
    if isinstance(xdata, np.ndarray) and isinstance(ydata, np.ndarray):
        return build_online_dataset(xdata, ydata, sequence_lengths,
                chunk_size, normalize_y = False, task_type = "classification")
    raise ValueError("Unexpected argument types to build_regression_dataset.")




def build_online_dataset(xdata, ydata, sequence_lengths = None,
        chunk_size:int = 2000, normalize_y:bool = True,
        task_type:str = "regression"):
    """build_online_dataset constructs an OnlineDataset
    object for data stored in memory, after first checking
    that some validity requirements are satisfied.

    Args:
        xdata (np.ndarray): A numpy array containing the x-values.
        ydata (np.ndarray): A numpy array containing the y-values.
        sequence_lengths (np.ndarray): Either None or a numpy array containing
            the length of each sequence / timeseries / graph for each datapoint.
            If you are not using a sequence kernel, this must always be None. If you are
            using a sequence / graph kernel, this cannot be None.
        chunk_size (int): The maximum size of data chunks that
            will be returned to callers. Limits memory consumption.
            Defaults to 2000.
        normalize_y (bool): If True, y values are normalized. Generally a
            good idea, unless you have already selected hyperparameters based
            on prior knowledge.
        task_type (str): One of "regression", "classification". Indicates
            what purpose the dataset will be used for.

    Returns:
        dataset (OnlineDataset): An object of class OnlineDataset
            that can be passed to the hyperparameter tuning
            and fitting routines of the model classes.

    Raises:
        ValueError: If the data passed is not valid, the function
            will raise a detailed ValueError explaining the issue.
    """
    if not isinstance(xdata, np.ndarray) or not isinstance(ydata, np.ndarray):
        raise ValueError("X and y must be numpy arrays!")

    __check_sequence_length(sequence_lengths, xdata, ydata)

    if len(ydata.shape) != 1:
        raise ValueError("Y must be a 1d numpy array.")
    if xdata.dtype not in ("float64", "float32"):
        raise ValueError("x must be an array of type float32 or type float64.")
    if ydata.dtype != "float64" and task_type == "regression":
        raise ValueError("For regression, ydata must be an array of type float64.")
    if task_type == "classification" and not issubclass(ydata.dtype.type,
                np.integer):
        raise ValueError("For classification, ydata must be an array of integers.")

    if ydata.shape[0] != xdata.shape[0]:
        raise ValueError("Different number of datapoints in x and y.")
    if np.isnan(xdata).any():
        raise ValueError("One or more elements in x is nan!")
    if np.max(xdata) > 1e15 or np.min(xdata) < -1e15:
        raise ValueError("Values > 1e15 or < -1e15 encountered. "
                    "Please rescale your data and check for np.inf.")

    if task_type == "regression":
        if normalize_y:
            dataset = OnlineDataset(xdata, ydata, sequence_lengths,
                chunk_size = chunk_size, trainy_mean = ydata.mean(),
                trainy_std = ydata.std())
        else:
            dataset = OnlineDataset(xdata, ydata, sequence_lengths,
                    chunk_size = chunk_size)

    else:
        dataset = OnlineDataset(xdata, ydata, sequence_lengths,
                chunk_size = chunk_size, max_class = ydata.max())
        if ydata.min() != 0:
            raise ValueError("For classification, there must be a zero category.")

    return dataset


def build_offline_np_dataset(xlist:list, ylist:list, sequence_lengths,
        chunk_size:int = 2000, normalize_y:bool = True,
        skip_safety_checks:bool = False, task_type:str = "regression"):
    """Constructs an OfflineDataset for data stored on disk
    as a list of npy files, after checking validity requirements.

    Args:
        xlist (list): A list of valid filepaths to .npy numpy
            arrays containing the x-data. These must all be either
            2d arrays where shape[1] is the number of features or
            3d arrays where shape[1] is the number of timepoints
            or sequence elements and shape[2] is the number of
            features for each timepoint / sequence element. They
            should be one or the other, not both.
        ylist (list): A list of valid filepaths to .npy numpy
            arrays containing the y-data. These must all be 1d
            arrays. The number of datapoints in each should
            match the corresponding element of xlist.
        sequence_lengths (np.ndarray): Either None or a list of valid filepaths to
            .npy files containing the length of each sequence / timeseries / graph. If
            not using a sequence kernel, this must always be None. If you are
            using a sequence / graph kernel, this cannot be None.
        chunk_size (int): The maximum size of data chunks that
            will be returned to callers. Limits memory consumption.
            Defaults to 2000.
        normalize_y (bool): If True, y values are normalized. Generally a
            good idea, unless you have already selected hyperparameters based
            on prior knowledge.
        skip_safety_checks (bool): If False, the builder will check
            each input array to make sure it does not contain
            infinite values or nan and that all the input arrays are 2d
            This is important -- if there are unexpected 3d arrays, np.nan,
            etc., this can lead to weird and unexpected results during
            fitting. On the other hand,
            this is a slow step and for a large dataset can take some
            time. Only skip_safety_checks if you have already checked your
            data and are confident that all is in order.
        task_type (str): One of "regression", "classification". Indicates
            what purpose the dataset will be used for.

    Returns:
        dataset (OfflineDataset): An object of class OfflineDataset
            that can be passed to the hyperparameter tuning
            and fitting routines of the model classes.

    Raises:
        ValueError: If the data passed is not valid, the function
            will raise a detailed ValueError explaining the issue.
    """
    if len(xlist) == 0:
        raise ValueError("At least one datafile must be supplied.")
    if len(xlist) != len(ylist):
        raise ValueError("xlist and ylist must have the same length.")
    if not isinstance(xlist, list) or not isinstance(ylist, list):
        raise ValueError("Both xlist and ylist should be lists.")

    if sequence_lengths is None:
        length_files = [None for y in ylist]
    else:
        length_files = sequence_lengths
        if len(sequence_lengths) != len(ylist):
            raise ValueError("sequence_lengths must either be None or "
                    "have the same length as ylist.")

    xshape = _get_array_file_shape(xlist[0])
    expected_arrlen = len(xshape)
    if expected_arrlen == 2:
        xdim = [0,-1]
    elif expected_arrlen == 3:
        xdim = [0,-1,-1]
    else:
        raise ValueError("Arrays should be either 2d or 3d.")

    if not skip_safety_checks:
        for xfile, yfile, length_file in zip(xlist, ylist, length_files):
            x_data = np.load(xfile)
            if x_data.shape[0] == 0:
                raise ValueError(f"File {xfile} has no datapoints.")
            if np.isnan(x_data).any():
                raise ValueError(f"One or more elements in file {xfile} is nan.")
            if np.max(x_data) > 1e15 or np.min(x_data) < -1e15:
                raise ValueError(f"One or more values in {xfile} is "
                        "> 1e15 or < -1e15. Please check for inf values "
                        "and / or rescale your data.")
            if len(x_data.shape) != expected_arrlen:
                raise ValueError(f"File {xfile} is not a {expected_arrlen}d "
                    "array, unlike some other arrays in xlist.")

            xdim[0] += x_data.shape[0]
            if xdim[1] == -1:
                xdim[1] = x_data.shape[1]
            if expected_arrlen == 2:
                if x_data.shape[1] != xdim[1]:
                    raise ValueError("All x arrays must have the same dimensionality.")
            elif expected_arrlen == 3:
                if xdim[2] == -1:
                    xdim[2] = x_data.shape[2]
                elif x_data.shape[2] != xdim[2]:
                    raise ValueError("All x arrays must have the same dimensionality.")

            ydata = np.load(yfile)
            if x_data.shape[0] != ydata.shape[0]:
                raise ValueError(f"File {xfile} has a different number of datapoints "
                    f"than file {yfile}.")
            if x_data.shape[0] > chunk_size:
                raise ValueError(f"Xfile {xfile} has more datapoints than allowed "
                    "based on specified chunk_size. Either increase chunk_size "
                    "or divide your data into np.ndarrays saved on disk that each "
                    "contain < chunk_size datapoints.")
            if len(ydata.shape) > 1:
                raise ValueError(f"The y file {yfile} is not a 1d array.")

            if length_file is None:
                __check_sequence_length(length_file, x_data, ydata)
            else:
                __check_sequence_length(np.load(length_file), x_data, ydata)


    #If we ARE skipping safety checks, retrieve the dimensionality info we will
    #need by checking files without loading them.
    else:
        xshape = _get_array_file_shape(xlist[0])
        xdim[1] = xshape[1]
        xdim[0] += xshape[0]
        if expected_arrlen == 3:
            xdim[2] = xshape[2]

    if normalize_y and task_type == "regression":
        max_class = 1
        trainy_mean, trainy_std = _get_offline_scaling_factors(ylist)
    else:
        trainy_mean, trainy_std = 0.0, 1.0
    if task_type == "classification":
        max_class, class_data_err = _get_offline_ymax(ylist)
        if class_data_err:
            raise ValueError("For classification, there must be a zero category, "
                "and all yfiles must be integers.")

    dataset = OfflineDataset(xlist, ylist, sequence_lengths, tuple(xdim),
                trainy_mean = trainy_mean, trainy_std = trainy_std,
                max_class = max_class, chunk_size = chunk_size)
    return dataset



def __check_sequence_length(seqlength, xdata, ydata):
    """Check a sequence length array to make sure it complies
    with typical requirements.

    Args:
        seqlength: Either None or a numpy array.
        ydata (np.ndarray): A numpy array.
        xdata (np.ndarray): A numpy array.

    Raises:
        ValueError: A ValueError is raised if requirements are
            not met. Otherwise if nothing is returned everything is ok.
    """
    if seqlength is not None:
        if len(xdata.shape) != 3:
            raise ValueError("sequence_length must be None if using "
                    "fixed vector input.")
        if not isinstance(seqlength, np.ndarray):
            raise ValueError("sequence_length must either be None or "
                    "a numpy array.")
        if not len(seqlength.shape) == 1:
            raise ValueError("sequence_length, if not None, must be a "
                    "1d numpy array.")
        if seqlength.shape[0] != ydata.shape[0]:
            raise ValueError("sequence_length, if not None, must have "
                    "the same length as ydata.")
        if not issubclass(seqlength.dtype.type, np.integer):
            raise ValueError("sequence_length, if not None, must be all "
                    "integers.")
        if seqlength.min() <= 0 or seqlength.max() > xdata.shape[1]:
            raise ValueError("sequence_length values must be in the range "
                    "(1, num_elements) where num_elements is the number of "
                    "sequence elements / graph nodes in the corresponding "
                    "input array.")
    elif len(xdata.shape) == 3:
        raise ValueError("sequence_lengths cannot be None if supplying "
                "sequences / time series / graphs as input.")


def _get_offline_ymax(yfiles:list):
    """Gets the max category for 'offline' data stored on disk.

    Args:
        yfiles (list): A list of valid filepaths for
            .npy files (numpy arrays) containing the
            y-values for all datapoints.

    Returns:
        max_class (int): The maximum category found in the data.
        class_data_err (bool): If True, one or more of the files
            has an incorrect format (noninteger data) or there
            is no zero category.
    """
    max_class, min_class, class_data_err = 0, 1, False
    for yfile in yfiles:
        y_data = np.load(yfile)
        if not issubclass(y_data.dtype.type, np.integer):
            class_data_err = True
            break
        if y_data.max() > max_class:
            max_class = y_data.max()
        if y_data.min() < min_class:
            min_class = y_data.min()

    if max_class == 0 or min_class != 0:
        class_data_err = True
    return max_class, class_data_err




def _get_offline_scaling_factors(yfiles:list):
    """Gets scaling factors (mean and standard deviation)
    for 'offline' data stored on disk.

    Args:
        yfiles (list): A list of valid filepaths for
            .npy files (numpy arrays) containing the
            y-values for all datapoints.

    Returns:
        trainy_mean (float): The mean of the training y-data.
        trainy_std (float): The standard deviation of the
            training y-data.
    """
    ndpoints = 0
    trainy_mean, trainy_var = 0.0, 0.0
    for yfile in yfiles:
        y_data = np.load(yfile).astype(np.float64)
        w_1 = y_data.shape[0] / (y_data.shape[0] + ndpoints)
        w_2 = ndpoints / (ndpoints + y_data.shape[0])
        w_3 = y_data.shape[0] * ndpoints / (y_data.shape[0] + ndpoints)**2
        y_data_mean = y_data.mean()

        trainy_var = w_1 * y_data.std()**2 + w_2 * trainy_var + w_3 * (y_data_mean -
                trainy_mean)**2
        trainy_mean = w_1 * y_data_mean + w_2 * trainy_mean
        ndpoints += y_data.shape[0]

    trainy_std = np.sqrt(trainy_var)
    return trainy_mean, trainy_std


def _get_array_file_shape(npy_file:str):
    """Gets the shape of a .npy file array without loading it
    to memory."""
    with open(npy_file, 'rb') as f_handle:
        major, _ = np.lib.format.read_magic(f_handle)
        if major == 1:
            arr_shape, _, _ = np.lib.format.read_array_header_1_0(f_handle)
        else:
            arr_shape, _, _ = np.lib.format.read_array_header_2_0(f_handle)
    return arr_shape



"""Contains some simple functions for generating probe
vectors for estimating NMLL."""
import numpy as np
try:
    import cupy as cp
except:
    pass

def generate_normal_probes_gpu(nsamples, num_rffs,
                    random_seed = 123,
                    preconditioner = None):
    """Generates probe vectors which may optionally be
    preconditioned (as preferred by caller) using a
    standard normal distribution.

    Args:
        nsamples (int): The number of probe vectors to generate.
        num_rffs (int): The number of random features.
        random_seed (int): A seed for the random number generator.
        preconditioner: Either None or a valid Preconditioner object.

    Returns:
        probes (cp.ndarray): An (num_rffs, nsamples) array of
            probe vectors for trace estimation.
    """
    rng = np.random.default_rng(random_seed)
    probes = rng.standard_normal(size=(num_rffs, nsamples))
    probes = cp.asarray(probes)
    if preconditioner is not None:
        probes = preconditioner.matvec_for_sampling(probes)
    return probes


def generate_radem_probes_gpu(nsamples, num_rffs,
                    random_seed = 123):
    """Generates probe vectors from a rademacher distribution.
    Args:
        nsamples (int): The number of probe vectors to generate.
        num_rffs (int): The number of random features.
        random_seed (int): A seed for the random number generator.

    Returns:
        probes (cp.ndarray): An (num_rffs, nsamples) array of
            probe vectors for trace estimation.
    """
    rng = np.random.default_rng(random_seed)
    radem = np.asarray([-1.0,1.0])
    probes = rng.choice(radem, size=(num_rffs, nsamples),
                    replace=True)
    probes /= np.linalg.norm(probes, axis=0)[None,:]
    return cp.asarray(probes)


def generate_normal_probes_cpu(nsamples, num_rffs,
                    random_seed = 123,
                    preconditioner = None):
    """Generates probe vectors which may optionally be
    preconditioned (as preferred by caller) using a
    standard normal distribution.

    Args:
        nsamples (int): The number of probe vectors to generate.
        num_rffs (int): The number of random features.
        random_seed (int): A seed for the random number generator.
        preconditioner: Either None or a valid Preconditioner object.

    Returns:
        probes (cp.ndarray): An (num_rffs, nsamples) array of
            probe vectors for trace estimation.
    """
    rng = np.random.default_rng(random_seed)
    probes = rng.standard_normal(size=(num_rffs, nsamples))
    if preconditioner is not None:
        probes = preconditioner.matvec_for_sampling(probes)
    return probes


def generate_radem_probes_cpu(nsamples, num_rffs,
                    random_seed = 123):
    """Generates probe vectors from a rademacher distribution.

    Args:
        nsamples (int): The number of probe vectors to generate.
        num_rffs (int): The number of random features.
        random_seed (int): A seed for the random number generator.

    Returns:
        probes (cp.ndarray): An (num_rffs, nsamples) array of
            probe vectors for trace estimation.
    """
    rng = np.random.default_rng(random_seed)
    radem = np.asarray([-1.0,1.0])
    probes = rng.choice(radem, size=(num_rffs, nsamples),
                        replace=True)
    probes /= np.linalg.norm(probes, axis=0)[None,:]
    return probes



"""A quick and dirty timeit script for checking the time to perform various
FHT operations and compare with cupy / numpy."""
import timeit
import numpy as np
import cupy as cp
import cupyx
from scipy.fftpack import dct
from xGPR.xgpr_cpu_rfgen_cpp_ext import cpuFastHadamardTransform as cFHT
from xGPR.xgpr_cpu_rfgen_cpp_ext import cpuSORFTransform as cSORF


#Running the CPU numpy test with too many columns is extremely
#expensive -- avoid running that test by accident.
MAX_COLS_MATMUL = 124#2048


def timetest():
    """Run a series of timeit tests and print the results."""
    #Change these lines depending on what dims you want to test.
    nthreads = 2
    nrows, ncols, nfeats = 2000, 1024, 2048
    if nfeats % ncols != 0:
        raise ValueError("nfeats must be an integer multiple of ncols.")

    nblocks = int(nfeats / ncols)
    ntests = 100
    print("**************************************")
    print(f"Running test with {nrows} rows, {ncols} cols, {nfeats} features")
    print("**************************************")
    matmul_setup = f"""
import numpy as np
random_seed = 123
rng = np.random.default_rng(random_seed)
marr = rng.uniform(low=-10.0, high=10.0, size=({nrows},{ncols}))
Q = rng.uniform(low=-1.0, high=1.0, size=({ncols},{nfeats}))
from __main__ import matmul_test"""
    if ncols < MAX_COLS_MATMUL:
        print("Time(us) for simple matrix multiplication:")
        time_taken = timeit.timeit("matmul_test(marr, Q)", setup=matmul_setup,
                number=ntests)
        print(1e6 * time_taken / ntests)


    fh3d_setup = f"""
nthreads = {nthreads}
import math
import numpy as np
from scipy.linalg import hadamard
from cpu_rf_gen_module import cpuFastHadamardTransform as cFHT
from cpu_rf_gen_module import cpuSORFTransform as cSORF
random_seed = 123
rng = np.random.default_rng(random_seed)
marr = rng.uniform(low=-10.0, high=10.0, size=({nrows},{nblocks},{ncols}))
radem_array = np.asarray([-1.0, 1.0])
D1 = rng.choice(radem_array, size=(3,{nblocks},{ncols}), replace=True)
from __main__ import fh3d_test"""
    print(f"Time(us) for fh3d version with {nthreads} threads:")
    time_taken = timeit.timeit("fh3d_test(marr, D1, nthreads)", setup=fh3d_setup,
                number=ntests)
    print(1e6 * time_taken / ntests)


    block_setup = f"""
nthreads = {nthreads}
import numpy as np
from scipy.linalg import hadamard
from cpu_rf_gen_module import cpuFastHadamardTransform as cFHT
from cpu_rf_gen_module import cpuSORFTransform as cSORF
random_seed = 123
rng = np.random.default_rng(random_seed)
marr = rng.uniform(low=-10.0, high=10.0, size=({nrows},{nblocks},{ncols}))
radem_array = np.asarray([-1, 1])
D1 = rng.choice(radem_array, size=(3,{nblocks},{ncols}), replace=True).astype(np.int8)
from __main__ import sorf_test"""
    print(f"Time(us) for block version with {nthreads} threads::")
    time_taken = timeit.timeit("sorf_test(marr, D1, nthreads)", setup=block_setup,
                number=ntests)
    print(1e6 * time_taken / ntests)


    '''random_seed = 123
    rng = np.random.default_rng(random_seed)
    marr = rng.uniform(low=-10.0, high=10.0, size=(nrows,nblocks,ncols))
    radem_array = np.asarray([-1, 1])
    diag1 = rng.choice(radem_array, size=(3,nblocks,ncols),
                    replace=True).astype(np.int8)
    marr = cp.asarray(marr).astype(cp.float32)
    diag1 = cp.asarray(diag1)
    print("Time for cuda version:")
    time_taken = cupyx.time.repeat(cuda_test, (marr, diag1), n_repeat = ntests)
    print(time_taken)'''


    dct_setup = f"""
import numpy as np, cupy as cp
random_seed = 123
rng = np.random.default_rng(random_seed)
marr = rng.uniform(low=-10.0, high=10.0, size=({nrows},{nblocks},{ncols}))
radem_array = np.asarray([-1, 1])
D1 = rng.choice(radem_array, size=(3,{nblocks},{ncols}), replace=True).astype(np.int8)
from __main__ import dct_test
from scipy.fftpack import dct"""
    print("Time(us) for DCT version:")
    time_taken = timeit.timeit("dct_test(marr, D1)", setup=dct_setup,
                number=ntests)
    print(1e6 * time_taken / ntests)

    random_seed = 123
    rng = np.random.default_rng(random_seed)
    marr = rng.uniform(low=-10.0, high=10.0, size=(nrows,ncols))
    qmat = rng.uniform(low=-1.0, high=1.0, size=(ncols,nfeats))
    qmat = cp.asarray(qmat).astype(cp.float32)
    marr = cp.asarray(marr).astype(cp.float32)
    print("Time for matrix multiplication with Cupy:")
    time_taken = cupyx.time.repeat(cp_matmul_test, (marr, qmat), n_repeat = ntests)
    print(time_taken)





def fh3d_test(marr, diag, nthreads):
    """Generate SORF features using the FHT module with
    separate diag matmul / fht operations."""
    marr = marr * diag[0:1,:,:]
    cFHT(marr, nthreads)
    marr = marr * diag[1:2,:,:]
    cFHT(marr, nthreads)
    marr = marr * diag[2:3,:,:]
    cFHT(marr, nthreads)


def sorf_test(marr, diag, nthreads):
    """Generate SORF features using the sorf function of
    the module."""
    cSORF(marr, diag, nthreads)


def matmul_test(marr, qmat):
    """Ordinary matrix multiplication."""
    _ = marr @ qmat


def cp_matmul_test(marr, qmat):
    """Matrix multiplication on CUDA."""
    _ = cp.matmul(marr, qmat)


def dct_test(marr, diag):
    """How fast would Scipy's DCT be by comparison?"""
    marr = marr * diag[0:1,:,:]
    marr = dct(marr, axis=-1, norm="ortho")
    marr = marr * diag[1:2,:,:]
    marr = dct(marr, axis=-1, norm="ortho")
    marr = marr * diag[2:3,:,:]
    marr = dct(marr, axis=-1, norm="ortho")

if __name__ == "__main__":
    timetest()



"""Tests the 'basic' hadamard transform based operations (SORF, SRHT, FHT)
for CPU and -- if GPU is available -- for GPU as well. We must test
for both single and double precision."""
import sys
import unittest
import numpy as np
from scipy.linalg import hadamard
import cupy as cp

from xGPR.xgpr_cpu_rfgen_cpp_ext import cpuFastHadamardTransform as cFHT
from xGPR.xgpr_cpu_rfgen_cpp_ext import cpuFastHadamardTransform2D as cFHT2D
from xGPR.xgpr_cpu_rfgen_cpp_ext import cpuSRHT as cSRHT

from xGPR.xgpr_cuda_rfgen_cpp_ext import cudaSRHT
from xGPR.xgpr_cuda_rfgen_cpp_ext import cudaFastHadamardTransform2D as cudaFHT2D


#To test the C extension for the fast hadamard transform, we compare the
#results to matrix multiplication with Hadamard matrices generated by
#Scipy (result should be the same).
class TestFastHadamardTransform(unittest.TestCase):
    """Runs tests for basic functionality (i.e. FHT and SORF)
    for float and double precision for CPU and (if available) GPU."""


    def test_3d_array_transform(self):
        """Tests the 3d FHT transform. This is for CPU only;
        the cuda module does not provide a separate function
        for FHT only at this time."""
        dim = (124,36,4)
        outcome_f, outcome_d = run_fht_test(dim)
        self.assertTrue(outcome_d)
        self.assertTrue(outcome_f)

        dim = (124,36,2)
        outcome_f, outcome_d = run_fht_test(dim)
        self.assertTrue(outcome_d)
        self.assertTrue(outcome_f)

        dim = (124,36,32)
        outcome_f, outcome_d = run_fht_test(dim)
        self.assertTrue(outcome_d)
        self.assertTrue(outcome_f)

        dim = (3001, 5, 1024)
        outcome_f, outcome_d = run_fht_test(dim)
        self.assertTrue(outcome_d)
        self.assertTrue(outcome_f)

        dim = (250, 1, 4096)
        outcome_f, outcome_d = run_fht_test(dim)
        self.assertTrue(outcome_d)
        self.assertTrue(outcome_f)


    def test_2d_array_transform(self):
        """Tests the 2d FHT transform."""
        dim = (124, 4)
        outcomes = run_fht_2d_test(dim)
        for outcome in outcomes:
            self.assertTrue(outcome)

        dim = (124, 2)
        outcomes = run_fht_2d_test(dim)
        for outcome in outcomes:
            self.assertTrue(outcome)

        dim = (124, 32)
        outcomes = run_fht_2d_test(dim)
        for outcome in outcomes:
            self.assertTrue(outcome)

        dim = (3001, 1024)
        outcomes = run_fht_2d_test(dim)
        for outcome in outcomes:
            self.assertTrue(outcome)

        dim = (250, 4096)
        outcomes = run_fht_2d_test(dim)
        for outcome in outcomes:
            self.assertTrue(outcome)


    def test_srht(self):
        """Tests SRHT functionality. Note that this tests SRHT
        functionality by using FHT. Therefore if the FHT did not
        pass, this one will not either."""

        outcomes = run_srht_test((150,256), 128)
        for outcome in outcomes:
            self.assertTrue(outcome)

        outcomes = run_srht_test((304,512), 256)
        for outcome in outcomes:
            self.assertTrue(outcome)

        outcomes = run_srht_test((5,2048), 512)
        for outcome in outcomes:
            self.assertTrue(outcome)


def run_fht_test(dim, random_seed = 123):
    """A helper function that runs an FHT test with specified
    dimensionality on a 3d input array."""
    scipy_double, scipy_float, marr, marr_float = setup_fht_test(dim, random_seed)
    cFHT(marr, 1)
    cFHT(marr_float, 1)
    outcome_d = np.allclose(scipy_double, marr)
    outcome_f = np.allclose(scipy_float, marr_float, rtol=1e-4, atol=1e-4)
    print("Did the C extension provide the correct result for "
            f"a {dim} array for floats, doubles? {outcome_f},{outcome_d}")
    return outcome_d, outcome_f



def run_fht_2d_test(dim, random_seed = 123):
    """A helper function that runs an FHT 2d test with specified
    dimensionality."""
    scipy_double, scipy_float, marr, marr_float = setup_fht_2d_test(dim, random_seed)
    cupy_marr, cupy_marr_float = marr.copy(), marr_float.copy()
    cFHT2D(marr, 1)
    cFHT2D(marr_float, 1)
    outcome_d = np.allclose(scipy_double, marr)
    outcome_f = np.allclose(scipy_float, marr_float, rtol=1e-4, atol=1e-4)
    print("Did the C extension provide the correct result for "
            f"a {dim} 2d array for floats, doubles? {outcome_f},{outcome_d}")

    if "cupy" in sys.modules:
        cupy_marr, cupy_marr_float = cp.asarray(cupy_marr), cp.asarray(cupy_marr_float)
        cudaFHT2D(cupy_marr)
        cudaFHT2D(cupy_marr_float)
        outcome_cuda_d = np.allclose(scipy_double, cupy_marr)
        outcome_cuda_f = np.allclose(scipy_float, cupy_marr_float, rtol=1e-4, atol=1e-4)
        print("Did the cuda extension provide the correct result for "
            f"a {dim} 2d array for floats, doubles? {outcome_cuda_f},{outcome_cuda_d}")
        return outcome_d, outcome_f, outcome_cuda_d, outcome_cuda_f

    return outcome_d, outcome_f


def setup_fht_test(dim, random_seed = 123):
    """A helper function that builds a matrix to test on
    with specified dimensions and gets the 'true' result using
    Scipy's Hadamard matrix generator. This is a very slow
    way to do a transform -- we use it here only to cross-
    check the results from our routines. Generated for
    both single and double precision."""
    rng = np.random.default_rng(random_seed)
    marr = rng.uniform(low=-10.0,high=10.0, size=dim)

    hmat = hadamard(marr.shape[2])
    scipy_results = []
    for i in range(marr.shape[1]):
        scipy_transform = np.matmul(hmat, marr[:,i,:].T).T
        scipy_results.append(scipy_transform)
    scipy_double = np.stack(scipy_results, axis=1)

    marr_float = marr.copy().astype(np.float32)
    scipy_results = []
    for i in range(marr_float.shape[1]):
        scipy_transform = np.matmul(hmat, marr_float[:,i,:].T).T
        scipy_results.append(scipy_transform)
    scipy_float = np.stack(scipy_results, axis=1)
    return scipy_double, scipy_float, marr, marr_float

def setup_fht_2d_test(dim, random_seed = 123):
    """A helper function that builds a matrix to test on
    with specified dimensions and gets the 'true' result using
    Scipy's Hadamard matrix generator. This is a very slow
    way to do a transform -- we use it here only to cross-
    check the results from our routines. Generated for
    both single and double precision."""
    rng = np.random.default_rng(random_seed)
    marr = rng.uniform(low=-10.0,high=10.0, size=dim)

    hmat = hadamard(marr.shape[1])
    scipy_double = np.matmul(hmat, marr.T).T

    marr_float = marr.copy().astype(np.float32)
    scipy_float = np.matmul(hmat, marr_float.T).T
    return scipy_double, scipy_float, marr, marr_float

def run_srht_test(dim, compression_size, random_seed = 123):
    """A helper function that runs the SRHT test for
    specified input dimensions."""
    marr_gt_double, marr_test_double, marr_gt_float, marr_test_float,\
            radem, norm_constant = setup_srht_test(dim,
                                    compression_size, random_seed)

    if "cupy" in sys.modules:
        cuda_test_double = cp.asarray(marr_test_double)
        cuda_test_float = cp.asarray(marr_test_float)

    marr_gt_double = marr_gt_double * radem[None,:] * norm_constant
    cFHT2D(marr_gt_double, 2)

    marr_gt_float = marr_gt_float * radem[None,:] * norm_constant
    cFHT2D(marr_gt_float, 2)

    cSRHT(marr_test_double, radem, 2)
    cSRHT(marr_test_float, radem, 2)
    outcome_d = np.allclose(marr_gt_double, marr_test_double)
    outcome_f = np.allclose(marr_gt_float, marr_test_float, rtol=1e-3,
            atol=1e-5)
    print("Did the C extension provide the correct result for SRHT of "
            f"a {dim} 2d array for floats, doubles? {outcome_f},{outcome_d}")

    if "cupy" in sys.modules:
        radem = cp.asarray(radem)
        cudaSRHT(cuda_test_float, radem, 2)
        cudaSRHT(cuda_test_double, radem, 2)
        cuda_test_double = cp.asnumpy(cuda_test_double)
        cuda_test_float = cp.asnumpy(cuda_test_float)
        outcome_cuda_d = np.allclose(marr_gt_double, cuda_test_double)
        outcome_cuda_f = np.allclose(marr_gt_float, cuda_test_float,
                rtol=1e-3, atol=1e-5)
        print("Did the Cuda extension provide the correct result for SRHT of "
            f"a {dim} 2d array for floats, doubles? {outcome_cuda_f},{outcome_cuda_d}")
        return outcome_d, outcome_f, outcome_cuda_d, outcome_cuda_f

    return outcome_d, outcome_f



def setup_srht_test(dim, compression_size, random_seed = 123):
    """A helper function that builds the matrices required for
    the SRHT test, specified using the input dimensions."""
    radem_array = np.asarray([-1.0,1.0], dtype=np.int8)
    rng = np.random.default_rng(random_seed)

    marr_gt_double = rng.uniform(low=-10.0,high=10.0, size=dim)
    marr_test_double = marr_gt_double.copy()
    marr_gt_float = marr_gt_double.copy().astype(np.float32)
    marr_test_float = marr_gt_float.copy()

    radem = rng.choice(radem_array, size=(dim[1]), replace=True)
    norm_constant = np.log2(dim[1]) / 2
    norm_constant = 1 / (2**norm_constant)
    return marr_gt_double, marr_test_double, marr_gt_float, marr_test_float, \
            radem, norm_constant


if __name__ == "__main__":
    unittest.main()



"""The radial basis function -- the most generic and popular of all
kernels."""
import numpy as np
from .sorf_kernel_baseclass import SORFKernelBaseclass


class RBF(SORFKernelBaseclass):
    """The RBF is the classic kernel and a perennial favorite. A GP
    equipped with this functions as a high-dimensional smoother.
    This class inherits from SORFKernelBaseclass which in turn inherits
    from KernelBaseclass. Only attributes unique to this child are
    described in this docstring.

    Attributes:
        hyperparams (np.ndarray): This kernel has two
            hyperparameters: lambda_ (noise)
            and sigma (inverse mismatch tolerance).
    """

    def __init__(self, xdim, num_rffs, random_seed = 123,
                device = "cpu", num_threads = 2, double_precision = False,
                kernel_spec_parms = {}):
        """Constructor for RBF.

        Args:
            xdim (tuple): The dimensions of the input.
            num_rffs (int): The user-requested number of random Fourier features.
            random_seed (int): The seed to the random number generator.
            device (str): One of 'cpu', 'gpu'. Indicates the starting device.
            num_threads (int): The number of threads to use if running on CPU. If
                running on GPU, this is ignored.
            double_precision (bool): If True, generate random features in double precision.
                Otherwise, generate as single precision.
            kernel_spec_parms (dict): A dictionary of other kernel-specific settings.
        """
        super().__init__(num_rffs, xdim, num_threads,
                sine_cosine_kernel = True, random_seed = random_seed,
                double_precision = double_precision,
                kernel_spec_parms = kernel_spec_parms)
        self.hyperparams = np.ones((2))
        self.bounds = np.asarray([[1e-3,4], [1e-6, 1e2]])

        self.device = device



"""Stores a dict of all kernel names mapping each to a
class. Should be updated any time a new kernel is
added / any time one is removed."""
from .basic_kernels.matern import Matern
from .basic_kernels.rbf import RBF
from .basic_kernels.linear import Linear
from .basic_kernels.cauchy import Cauchy
from .basic_kernels.rbf_linear import RBFLinear

from .convolution_kernels.conv1d_rbf import Conv1dRBF
from .convolution_kernels.conv1d_matern import Conv1dMatern
from .convolution_kernels.conv1d_cauchy import Conv1dCauchy
from .convolution_kernels.l2_conv1d import Conv1dTwoLayer

from .convolution_kernels.graph_rbf import GraphRBF
from .convolution_kernels.graph_matern import GraphMatern
from .convolution_kernels.graph_cauchy import GraphCauchy

from .ARD_kernels.mini_ard import MiniARD


KERNEL_NAME_TO_CLASS = {"RBF":RBF,
        "Matern":Matern,
        "Cauchy":Cauchy,
        "Conv1dRBF":Conv1dRBF,
        "Conv1dMatern":Conv1dMatern,
        "Conv1dCauchy":Conv1dCauchy,
        "Conv1dTwoLayer":Conv1dTwoLayer,
        "GraphRBF":GraphRBF,
        "GraphMatern":GraphMatern,
        "GraphCauchy":GraphCauchy,
        "Linear":Linear,
        "RBFLinear":RBFLinear,
        "MiniARD":MiniARD}

# A list of kernels that require 3d arrays as input. This
# is used by kernel_fgen for generating random features
# outside of a regressor / classifier.
ARR_3D_KERNELS = {"GraphRBF", "Conv1dRBF", "Conv1dMatern",
        "GraphMatern", "GraphCauchy", "Conv1dCauchy",
        "Conv1dTwoLayer"}



"""Magic numbers for use by model classes."""
MAX_VARIANCE_RFFS = 4096
MAX_CLOSED_FORM_RFFS = 8192

#The following are default settings for kernels that have
#special parameters.
DEFAULT_KERNEL_SPEC_PARMS = {"matern_nu":5/2, "intercept":True,
        "averaging":"none"}

#The following is used if an error is encountered during
#matrix decomposition.
DEFAULT_SCORE_IF_PROBLEM = 1e40

#Default settings for NMLL approximation.
default_nmll_params = {"max_rank":1024, "preconditioner_mode":"srht_2",
        "nsamples":25, "nmll_iter":500, "nmll_tol":1e-6}

#Default max rank for NMLL approximation.
LARGEST_NMLL_MAX_RANK = 3000
#Default min rank for NMLL approximation.
SMALLEST_NMLL_MAX_RANK = 512



"""Implements a randomized nystrom preconditioner that can
be switched back and forth between Cuda and CPU. This is not
possible for the other variants since they have to inherit from
either Scipy or Cupy.scipy.sparse."""
import numpy as np
try:
    import cupy as cp
except:
    pass
from .rand_nys_constructors import initialize_srht, initialize_srht_multipass



class InterDevicePreconditioner():
    """Implements a preconditioner based on the randomized
    Nystrom approximation to the inverse of Z^T Z + lambda;
    unlike the other preconditioners, this one can be switched
    back and forth between devices.

    Attributes:
        achieved_ratio (float): lambda_ / min_eig, where min_eig is the
            smallest eigenvalue for the preconditioner. This is a
            reasonably good predictor of how effective the preconditioner
            will be. Smaller is better.
        prefactor (float): A constant by which the matvec is multiplied
            when performing matvecs with the preconditioner.
        u_mat (array): A cupy array containing the eigenvectors of
            the matrix formed during preconditioner construction. Used
            together with prefactor and inv_eig to perform the matvec
            that approximates A^-1.
        inv_eig (array): A cupy array containing the inverse of the
            eigenvalues of the matrix formed during preconditioner construction.
    """

    def __init__(self, kernel, dataset, max_rank, verbose,
                random_state = 123, method = "srht"):
        """Class constructor.

        Args:
            kernel: A valid kernel object that can generate random features.
            dataset: A valid online or offline dataset object that can retrieve
                chunked data.
            max_rank (int): The rank of the preconditioner.
            verbose (bool): If True, print regular updates.
            random_state (int): The random number seed for the random number
                generator.
            method (str): One of "srht", "gauss", "srht_2". Determines the method of
                preconditioner construction.
        """
        if method not in ["srht_2", "srht_3", "srht"]:
            raise ValueError("Unknown method supplied for preconditioner "
                    "construction.")

        if method.startswith("srht_"):
            n_passes = int(method.split("_")[1])
            self.u_mat, s_mat, _, _ = initialize_srht_multipass(dataset, max_rank,
                                kernel, random_state, verbose, n_passes)
        elif method == "srht":
            self.u_mat, s_mat, _, _ = initialize_srht(dataset, max_rank,
                                kernel, random_state, verbose)

        lambda_ = kernel.get_lambda()

        self.inv_eig = s_mat + lambda_**2
        mask = self.inv_eig > 1e-14
        self.inv_eig[mask] = 1 / self.inv_eig[mask]
        self.inv_eig[mask==False] = 0.0

        min_eig = s_mat.min()
        self.achieved_ratio = float(min_eig / lambda_**2)
        self.prefactor = float(min_eig + lambda_**2)

        self.device = kernel.device


    def batch_matvec(self, xvec):
        """Returns a matvec of the preconditioner with a set of input
        vectors xvec."""
        xprod = self.u_mat.T @ xvec
        xprod1 = self.u_mat @ (self.inv_eig[:,None] * self.prefactor * xprod)
        xprod2 = xvec - (self.u_mat @ xprod)
        return xprod2 + xprod1


    @property
    def device(self):
        """Getter for the device property, which determines
        whether calculations are on CPU or GPU."""
        return self.device_


    @device.setter
    def device(self, value):
        """Setter for device, which determines whether calculations
        are on CPU or GPU.

        Args:
            value (str): Must be one of 'cpu', 'cuda'.

        Raises:
            ValueError: A ValueError is raised if an unrecognized
                device is passed.
        """
        if value == "cpu":
            if not isinstance(self.u_mat, np.ndarray):
                self.u_mat = cp.asnumpy(self.u_mat)
                self.inv_eig = cp.asnumpy(self.inv_eig)

        elif value == "cuda":
            self.u_mat = cp.asarray(self.u_mat)
            self.inv_eig = cp.asarray(self.inv_eig)

        else:
            raise ValueError("Unrecognized device supplied. Must be one "
                    "of 'cpu', 'cuda'.")

        self.device_ = value



"""This module describes the OnlineDataset for regression and
classification.

The OnlineDataset bundles together attributes for handling
a dataset stored in memory. The dataset can return a specific
chunk of data from the list provided to it, can serve as a generator
and return all chunks in succession, or can provide a minibatch.
"""
import numpy as np

from .data_handling_baseclass import DatasetBaseclass


class OnlineDataset(DatasetBaseclass):
    """The OnlineDataset class handles datasets which are stored in
    memory. Only unique attributes not shared by parent class are
    described here.

    Attributes:
        _xdata (array): A cupy or numpy array containing the x data.
        _ydata (array): A cupy or numpy array containing the y data.
        _sequence_lengths: Either None or a cupy / numpy array containing
            the sequence lengths (for graph / sequence kernels).
    """
    def __init__(self, xdata, ydata, sequence_lengths = None,
                       chunk_size = 2000,
                       trainy_mean = 0.,
                       trainy_std = 1.,
                       max_class = 1.):
        """Constructor for the OnlineDataset class.

        Args:
            xdata (array): A numpy array containing the x data.
            ydata (array): A numpy array containing the y data.
            device (str): The current device. Must be in ["cpu", "gpu"].
            chunk_size (int): When returning chunks of data, the chunks
                are of size chunk_size. This limits memory consumption
                (by avoiding situations where we try to featurize
                too many datapoints at once).
            trainy_mean (float): The mean of the y-values. Only used for
                regression.
            trainy_std (float): The standard deviation of the y-values.
                Only used for regression.
            max_class (int): The largest category number in the data. Only
                used for classification.
        """
        super().__init__(xdata.shape, chunk_size, trainy_mean,
                trainy_std, max_class)
        self._xdata = xdata
        self._ydata = ydata
        self._sequence_lengths = sequence_lengths


    def get_chunked_data(self):
        """A generator that returns the stored data in chunks
        of size chunk_size."""
        if self._sequence_lengths is None:
            for i in range(0, self._xdim[0], self.get_chunk_size()):
                cutoff = min(i + self.get_chunk_size(), self._xdim[0])
                xchunk = self._xdata[i:cutoff,...]
                ychunk = self._ydata[i:cutoff]
                ychunk = ychunk.astype(np.float64)
                ychunk -= self._trainy_mean
                ychunk /= self._trainy_std
                yield xchunk, ychunk, None

        else:
            for i in range(0, self._xdim[0], self.get_chunk_size()):
                cutoff = min(i + self.get_chunk_size(), self._xdim[0])
                xchunk = self._xdata[i:cutoff,...]
                ychunk = self._ydata[i:cutoff]
                lchunk = self._sequence_lengths[i:cutoff]

                ychunk = ychunk.astype(np.float64)
                ychunk -= self._trainy_mean
                ychunk /= self._trainy_std
                yield xchunk, ychunk, lchunk



    def get_chunked_x_data(self):
        """A generator that loops over the xdata only in chunks
        of size chunk_size."""
        if self._sequence_lengths is None:
            for i in range(0, self._xdim[0], self.get_chunk_size()):
                cutoff = min(i + self.get_chunk_size(), self._xdim[0])
                yield self._xdata[i:cutoff,...], None
        else:
            for i in range(0, self._xdim[0], self.get_chunk_size()):
                cutoff = min(i + self.get_chunk_size(), self._xdim[0])
                xchunk = self._xdata[i:cutoff,...]
                lchunk = self._sequence_lengths[i:cutoff]
                yield xchunk, lchunk



'''The RationalQuadratic kernel is a popular alternative to the RBF.'''
import numpy as np
from .sorf_kernel_baseclass import SORFKernelBaseclass



class Cauchy(SORFKernelBaseclass):
    """The rational quadratic kernel is a popular alternative to
    the RBF. Here it is implemented with small alpha (large alpha values
    are indistinguishable from an RBF and offer no advantage). This class
    inherits from SORFKernelBaseclass which in turn inherits from
    KernelBaseclass. Only attributes unique to this child are described
    in this docstring.

    Attributes:
        hyperparams (np.ndarray): This kernel has two
            hyperparameters: lambda_ (noise)
            and sigma (inverse mismatch tolerance).
    """

    def __init__(self, xdim, num_rffs, random_seed = 123, device = "cpu",
                num_threads = 2, double_precision = False,
                kernel_spec_parms = {}):
        """Constructor.

        Args:
            xdim (tuple): The dimensions of the input.
            num_rffs (int): The user-requested number of random Fourier features.
            random_seed (int): The seed to the random number generator.
            device (str): One of 'cpu', 'gpu'. Indicates the starting device.
            num_threads (int): The number of threads to use for generating random
                features if running on CPU. If running on GPU, this is ignored.
            num_threads (int): The number of threads to use for generating random
                features if running on CPU. If running on GPU, this is ignored.
            double_precision (bool): If True, generate random features in double precision.
                Otherwise, generate as single precision.
        """
        super().__init__(num_rffs, xdim, num_threads,
                sine_cosine_kernel = True, random_seed = random_seed,
                double_precision = double_precision,
                kernel_spec_parms = kernel_spec_parms)

        rng = np.random.default_rng(random_seed)
        dstsamples = np.sqrt(rng.exponential(size=self.num_freqs))
        self.chi_arr *= dstsamples

        self.hyperparams = np.ones((2))
        self.bounds = np.asarray([[1e-3,1e1], [1e-6, 1e2]])
        self.device = device



"""Tests the RBF feature generation routines (specific for RBF, Matern
and MiniARD / ARD kernels, which by extension includes static layer kernels."""
import sys
import unittest
from math import ceil
import numpy as np
from scipy.stats import chi
import cupy as cp

from xGPR.xgpr_cpu_rfgen_cpp_ext import cpuRBFFeatureGen as cRBF
from xGPR.xgpr_cpu_rfgen_cpp_ext import cpuRBFGrad as cRBFGrad
from xGPR.xgpr_cpu_rfgen_cpp_ext import cpuFastHadamardTransform2D as cFHT

from xGPR.xgpr_cuda_rfgen_cpp_ext import cudaRBFFeatureGen as cudaRBF
from xGPR.xgpr_cuda_rfgen_cpp_ext import cudaRBFGrad


class TestRBFFeatureGen(unittest.TestCase):
    """Runs tests for RBF feature generation for float and double
    precision for CPU and (if available) GPU. We do this by comparing
    with feature generation using a Python routine that
    makes use of SORFTransform functions, so the basic_fht test must
    pass in order for this test to work."""


    def test_rbf_feature_gen(self):
        """Tests RBF feature generation for CPU and if available GPU."""
        for n_freqs in [64, 2000, 8192]:
            outcomes = run_rbf_test((10,50), n_freqs)
            for outcome in outcomes:
                self.assertTrue(outcome)

            outcomes = run_rbf_test((10,3), n_freqs)
            for outcome in outcomes:
                self.assertTrue(outcome)

            outcomes = run_rbf_test((3,2003), n_freqs)
            for outcome in outcomes:
                self.assertTrue(outcome)

            outcomes = run_rbf_test((11,1076), n_freqs)
            for outcome in outcomes:
                self.assertTrue(outcome)

            outcomes = run_rbf_test((231,856), n_freqs, fit_intercept = True)
            for outcome in outcomes:
                self.assertTrue(outcome)



    def test_rbf_grad_calc(self):
        """Tests RBF gradient calc for CPU and if available GPU."""
        for n_freqs in [64, 2000, 8192]:
            outcomes = run_rbf_grad_test((10,50), n_freqs)
            for outcome in outcomes:
                self.assertTrue(outcome)

            outcomes = run_rbf_grad_test((513,232), n_freqs, fit_intercept = True)
            for outcome in outcomes:
                self.assertTrue(outcome)

            outcomes = run_rbf_grad_test((11,3001), n_freqs)
            for outcome in outcomes:
                self.assertTrue(outcome)

            outcomes = run_rbf_grad_test((11,3), n_freqs)
            for outcome in outcomes:
                self.assertTrue(outcome)


def run_rbf_test(xdim, num_freqs, random_seed = 123, fit_intercept = False):
    """A helper function that runs the RBF test for
    specified input dimensions."""

    test_array, radem, chi_arr, nblocks, padded_dims = \
            setup_rbf_test(xdim, num_freqs, random_seed)

    gt_double, _ = generate_rbf_values(test_array, radem, chi_arr,
            nblocks, padded_dims, fit_intercept)
    gt_float, _ = generate_rbf_values(test_array.astype(np.float32), radem,
            chi_arr.astype(np.float32), nblocks, padded_dims, fit_intercept)

    double_output = np.zeros((test_array.shape[0], num_freqs * 2))
    cRBF(test_array, double_output, radem, chi_arr, 2, fit_intercept)

    float_output = np.zeros((test_array.shape[0], num_freqs * 2))
    cRBF(test_array.astype(np.float32), float_output, radem,
            chi_arr.astype(np.float32), 2, fit_intercept)

    if "cupy" in sys.modules:
        cuda_test_array = cp.asarray(test_array)
        radem = cp.asarray(radem)
        chi_arr = cp.asarray(chi_arr)
        cuda_double_output = cp.zeros((test_array.shape[0], num_freqs * 2))
        cuda_float_output = cp.zeros((test_array.shape[0], num_freqs * 2))

        cudaRBF(cuda_test_array, cuda_double_output, radem,
                chi_arr, fit_intercept)
        cudaRBF(cuda_test_array.astype(cp.float32), cuda_float_output, radem,
                chi_arr.astype(cp.float32), fit_intercept)


    outcome_d = np.allclose(gt_double, double_output)
    outcome_f = np.allclose(gt_float, float_output, rtol=1e-5, atol=1e-4)
    print("Correct result for CPU  for RBF of "
            f"{xdim}, {num_freqs} for float, double? {outcome_f},{outcome_d}")

    if "cupy" in sys.modules:
        outcome_cuda_d = np.allclose(gt_double, cuda_double_output)
        outcome_cuda_f = np.allclose(gt_float, cuda_float_output, rtol=1e-5,
                atol=1e-4)
        print("Correct result for Cuda for RBF of "
            f"{xdim}, {num_freqs} for float, double? {outcome_cuda_f},{outcome_cuda_d}")
        return outcome_d, outcome_f, outcome_cuda_d, outcome_cuda_f
    return outcome_d, outcome_f




def run_rbf_grad_test(xdim, num_freqs, random_seed = 123,
        fit_intercept = False):
    """A helper function that tests the Cython wrapper which both
    generates RFs and calculates the gradient for specified input params."""

    test_array, radem, chi_arr, nblocks, padded_dims = \
            setup_rbf_test(xdim, num_freqs, random_seed)

    gt_double, gt_double_grad = generate_rbf_values(test_array, radem, chi_arr,
            nblocks, padded_dims, fit_intercept)
    gt_float, gt_float_grad = generate_rbf_values(test_array.astype(np.float32),
            radem, chi_arr.astype(np.float32), nblocks, padded_dims,
            fit_intercept)

    double_output = np.zeros((test_array.shape[0], num_freqs * 2))
    double_grad = np.zeros((double_output.shape[0], double_output.shape[1], 1))
    cRBFGrad(test_array, double_output, double_grad, radem, chi_arr,
            1.0, 2, fit_intercept)

    float_output = np.zeros((test_array.shape[0], num_freqs * 2))
    float_grad = np.zeros((float_output.shape[0], float_output.shape[1], 1))
    cRBFGrad(test_array.astype(np.float32), float_output,
            float_grad, radem, chi_arr.astype(np.float32), 1.0,
            2, fit_intercept)

    if "cupy" in sys.modules:
        cuda_test_array = cp.asarray(test_array)
        radem = cp.asarray(radem)
        chi_arr = cp.asarray(chi_arr)
        cuda_double_output = cp.zeros((test_array.shape[0], num_freqs * 2))
        cuda_float_output = cp.zeros((test_array.shape[0], num_freqs * 2))
        cuda_double_grad = cp.zeros((test_array.shape[0], num_freqs * 2, 1))
        cuda_float_grad = cp.zeros((test_array.shape[0], num_freqs * 2, 1))

        cudaRBFGrad(cuda_test_array, cuda_double_output, cuda_double_grad, radem,
                chi_arr, 1.0, fit_intercept)
        cudaRBFGrad(cuda_test_array.astype(cp.float32),
                cuda_float_output, cuda_float_grad, radem,
                chi_arr.astype(cp.float32), 1.0,
                fit_intercept)

    outcome_d = np.allclose(gt_double, double_output)
    outcome_f = np.allclose(gt_float, float_output, atol=1e-4,
            rtol=1e-4)
    outcome_grad_d = np.allclose(gt_double_grad, double_grad)
    outcome_grad_f = np.allclose(gt_float_grad, float_grad,
            atol=1e-4, rtol=1e-4)
    print("Correct result for CPU  for gradient RBF of "
            f"{xdim}, {num_freqs} float, double? {outcome_f},{outcome_d}")
    print("Correct result for Cuda for gradient RBF of "
            f"{xdim}, {num_freqs} float, double? "
            f"{outcome_grad_f},{outcome_grad_d}")

    if "cupy" in sys.modules:
        outcome_cuda_d = np.allclose(gt_double, cuda_double_output)
        outcome_cuda_f = np.allclose(gt_float, cuda_float_output,
                atol=1e-4, rtol=1e-4)
        outcome_cuda_grad_d = np.allclose(gt_double_grad, cuda_double_grad)
        outcome_cuda_grad_f = np.allclose(gt_float_grad, cuda_float_grad,
                atol=1e-2, rtol=1e-2)
        print("Correct result for CPU  for gradient RBF of "
            f"{xdim}, {num_freqs} float, double? {outcome_cuda_f},{outcome_cuda_d}")
        print("Correct result for Cuda for gradient RBF of "
            f"{xdim}, {num_freqs} float, double? "
            f"{outcome_cuda_grad_f},{outcome_cuda_grad_d}")
        return outcome_d, outcome_f, outcome_cuda_d, outcome_cuda_f, \
                outcome_grad_d, outcome_grad_f, outcome_cuda_grad_d, \
                outcome_cuda_grad_f
    return outcome_d, outcome_f, outcome_grad_d, outcome_grad_f



def setup_rbf_test(xdim, num_freqs, random_seed = 123):
    """A helper function that builds the matrices required for
    the RBF test, specified using the input dimensions."""
    padded_dims = 2**ceil(np.log2(max(xdim[-1], 2)))
    if padded_dims < num_freqs:
        nblocks = ceil(num_freqs / padded_dims)
    else:
        nblocks = 1

    radem_array = np.asarray([-1,1], dtype=np.int8)
    rng = np.random.default_rng(random_seed)


    radem = rng.choice(radem_array, size=(3,1,nblocks * padded_dims),
            replace=True)
    chi_arr = chi.rvs(df=padded_dims, size=num_freqs,
                            random_state = random_seed)

    test_array = rng.uniform(low=-10.0,high=10.0, size=(xdim[0], xdim[1]))
    return test_array, radem, chi_arr, nblocks, padded_dims


def generate_rbf_values(test_array, radem, chi_arr, nblocks,
        padded_dims, fit_intercept = False):
    """Generates the 'ground-truth' RBF values for
    comparison with those from the C / Cuda extension,
    using a very inefficient but easy to troubleshoot
    pure Python approach."""
    pretrans_x = []
    norm_constant = np.log2(padded_dims) / 2
    norm_constant = 1 / (2**norm_constant)

    for i in range(nblocks):
        temp_arr = np.zeros((test_array.shape[0], padded_dims),
                dtype = test_array.dtype)
        temp_arr[:,:test_array.shape[1]] = test_array

        for j in range(3):
            temp_arr *= radem[j, 0, i*padded_dims:(i+1)*padded_dims] * norm_constant
            cFHT(temp_arr, 1)

        #The simplex projection (Reid et al 2023) (disabled for now
        #since erratic / unclear effect on performance). We use a deliberately
        #clumsy / inefficient approach here to ensure that numpy
        #will use 32-bit float precision if the input is 32-bit
        #(otherwise numpy tends to default to 64-bit and the result
        #may not be np.allclose to the 32-bit c extension calculation)
        #scalar = np.sqrt(padded_dims - 1, dtype=temp_arr.dtype)
        #sum_arr = np.zeros((temp_arr.shape[0]), dtype=temp_arr.dtype)
        #for j in range(temp_arr.shape[1] - 1):
        #    sum_arr += temp_arr[:,j]
        #sum_arr /= scalar
        #temp_arr[:,-1] = sum_arr
        #scalar = ((1 + np.sqrt(padded_dims, dtype=temp_arr.dtype)) /
        #        (padded_dims - 1)).astype(temp_arr.dtype)
        #sum_arr *= scalar
        #scalar = np.sqrt(padded_dims / (padded_dims - 1),
        #        dtype=temp_arr.dtype)
        #temp_arr[:,:-1] = temp_arr[:,:-1] * scalar - sum_arr[:,None]

        pretrans_x.append(temp_arr)

    pretrans_x = np.hstack(pretrans_x)[:,:chi_arr.shape[0]]
    pretrans_x *= chi_arr[None,:]


    xtrans = np.zeros((test_array.shape[0], chi_arr.shape[0] * 2))
    gradient = np.zeros((test_array.shape[0], chi_arr.shape[0] * 2, 1))

    for j in range(0, chi_arr.shape[0], 1):
        xtrans[:,2*j] = np.cos(pretrans_x[:,j])
        xtrans[:,2*j+1] = np.sin(pretrans_x[:,j])

        gradient[:,2*j,0] = -xtrans[:,2*j+1] * \
                pretrans_x[:,j].astype(np.float64)
        gradient[:,2*j+1,0] = xtrans[:,2*j] * \
                pretrans_x[:,j].astype(np.float64)

    if fit_intercept:
        xtrans *= np.sqrt(1 / (chi_arr.shape[0]-0.5))
        gradient *= np.sqrt(1 / (chi_arr.shape[0]-0.5))
    else:
        xtrans *= np.sqrt(1 / chi_arr.shape[0])
        gradient *= np.sqrt(1 / chi_arr.shape[0])
    return xtrans, gradient


if __name__ == "__main__":
    unittest.main()



"""This kernel is equivalent to the Conv1dCauchy kernel but is specialized
to work on graphs, where we only ever want a convolution width of 1."""
import numpy as np
from .conv_kernel_baseclass import ConvKernelBaseclass


class GraphCauchy(ConvKernelBaseclass):
    """This is similar to sequence kernels but is specialized to work
    on graphs, where the input is a sequence of node descriptions.
    This class inherits from ConvKernelBaseclass.
    Only attributes unique to this child are described in this docstring.
    
    Attributes:
        hyperparams (np.ndarray): This kernel has two
            hyperparameters: lambda_ (noise)
            and sigma (inverse mismatch tolerance).
        bounds (np.ndarray): The builtin bounds for hyperparameter optimization,
            which can be overriden / reset by user.
    """

    def __init__(self, xdim, num_rffs, random_seed = 123, device = "cpu",
                    num_threads = 2, double_precision = False, kernel_spec_parms = {}):
        """Constructor.

        Args:
            xdim (tuple): The dimensions of the input. Either (N, D) or (N, M, D)
                where N is the number of datapoints, D is number of features
                and M is number of timepoints or sequence elements (convolution
                kernels only).
            num_rffs (int): The user-requested number of random Fourier features.
                For sine-cosine kernels (RBF, Cauchy), this will be saved by the
                class as num_rffs.
            random_seed (int): The seed to the random number generator.
            device (str): One of 'cpu', 'gpu'. Indicates the starting device.
            num_threads (int): The number of threads to use for random feature generation
                if running on CPU. Ignored if running on GPU.
            double_precision (bool): If True, generate random features in double precision.
                Otherwise, generate as single precision.

        Raises:
            ValueError: A ValueError is raised if the dimensions of the input are
                inappropriate given the conv_width.
        """
        super().__init__(xdim, num_rffs, random_seed,
                num_threads, double_precision, 1,
                kernel_spec_parms)
        rng = np.random.default_rng(random_seed)
        dstsamples = np.sqrt(rng.exponential(size=self.num_freqs))
        self.chi_arr *= dstsamples

        self.hyperparams = np.ones((2))
        self.bounds = np.asarray([[1e-3,1e1], [1e-6, 1e2]])
        self.device = device



'''The Matern kernel can handle "squiggly" or bumpy functions unlike the RBF,
but like the RBF is prone to "ringing" when discontinuities are encountered.
The closed form kernel is the Fourier transform of a student t distribution.'''
import numpy as np
from .sorf_kernel_baseclass import SORFKernelBaseclass



class Matern(SORFKernelBaseclass):
    """The Matern kernel often offers performance similar to RBF but
    sometimes with slight improvement owing to its greater
    ability to handle 'bumpy' functions. This class inherits from
    SORFKernelBaseclass which in turn inherits from KernelBaseclass.
    Only attributes unique to this child are described in this docstring.

    Attributes:
        hyperparams (np.ndarray): This kernel has two
            hyperparameters: lambda_ (noise)
            and sigma (inverse mismatch tolerance).
        matern_nu (float): The nu hyperparamter of the Matern kernel.
            Set by the user. Must be >= 1/2, <= 5/2. Larger values
            indicate smoother functions are expected. 5/2 is
            recommended.
    """

    def __init__(self, xdim, num_rffs, random_seed = 123, device = "cpu",
                num_threads = 2, double_precision = False,
                kernel_spec_parms = {}):
        """Constructor for Matern.

        Args:
            xdim (tuple): The dimensions of the input.
            num_rffs (int): The user-requested number of random Fourier features.
            random_seed (int): The seed to the random number generator.
            device (str): One of 'cpu', 'gpu'. Indicates the starting device.
            num_threads (int): The number of threads to use for generating random
                features if running on CPU. If running on GPU, this is ignored.
            num_threads (int): The number of threads to use for generating random
                features if running on CPU. If running on GPU, this is ignored.
            double_precision (bool): If True, generate random features in double precision.
                Otherwise, generate as single precision.
            kernel_spec_parms (dict): A dictionary of additional kernel-specific
                attributes. In this case, should contain 'matern_nu'.
        """
        super().__init__(num_rffs, xdim, num_threads,
                sine_cosine_kernel = True, random_seed = random_seed,
                double_precision = double_precision,
                kernel_spec_parms = kernel_spec_parms)
        if "matern_nu" not in kernel_spec_parms:
            raise ValueError("Tried to initialize a Matern kernel without supplying nu.")

        self.matern_nu = kernel_spec_parms["matern_nu"]
        if self.matern_nu < 1/2 or self.matern_nu > 5/2:
            raise ValueError("nu must be >= 1/2 and <= 5/2.")
        rng = np.random.default_rng(random_seed)
        chisamples = np.sqrt(rng.chisquare(2 * self.matern_nu,
                                    size=self.num_freqs)
                                    / (self.matern_nu * 2) )
        self.chi_arr /= chisamples

        self.hyperparams = np.ones((2))
        self.bounds = np.asarray([[1e-3,1e1], [1e-6, 1e2]])
        self.device = device



"""A linear kernel, which corresponds to Bayesian linear
regression."""
import numpy as np
try:
    import cupy as cp
except:
    pass
from ..kernel_baseclass import KernelBaseclass


class Linear(KernelBaseclass):
    """The Linear kernel corresponds to Bayesian linear regression.
    For attributes not described here, see the baseclass.
    """

    def __init__(self, xdim, num_rffs, random_seed = 123,
                device = "cpu", num_threads = 2,
                double_precision = True,
                kernel_spec_parms = {}):
        """Constructor.

        Args:
            xdim (tuple): The dimensions of the input.
            num_rffs (int): The user-requested number of random Fourier features.
                For this kernel, it is ignored.
            random_seed (int): The seed to the random number generator.
            device (str): One of 'cpu', 'gpu'. Indicates the starting device.
            num_threads (int): The number of threads to use for random feature generation
                if running on CPU; if running on GPU this is ignored. Since random features
                are not generated for this kernel this is ignored.
            double_precision (bool): Not used for this kernel; accepted to preserve
                common interface with other kernels.
            kernel_spec_parms (dict): A dictionary of kernel-specific parameters.
                In this case, may optionally contain 'intercept'; if false, no y-intercept
                is fitted.
        """
        self.fit_intercept = True
        actual_rffs = xdim[1] + 1
        if "intercept" in kernel_spec_parms:
            if kernel_spec_parms["intercept"] is False:
                self.fit_intercept = False
                actual_rffs = xdim[1]

        super().__init__(actual_rffs, xdim, kernel_spec_parms = kernel_spec_parms)

        if len(xdim) > 2:
            raise ValueError("The Linear kernel is only applicable for "
                    "fixed vector input.")
        self.hyperparams = np.ones((1))
        self.bounds = np.asarray([[1e-3,1e1]])

        self.device = device


    def kernel_specific_set_device(self, new_device):
        """Provided for consistency with baseclass. This
        kernel -- unlike all others -- has no device-specific
        params."""
        return


    def kernel_specific_set_hyperparams(self):
        """Provided for consistency with baseclass. This
        kernel has no kernel-specific properties that must
        be reset after hyperparameters are changed."""
        return


    def kernel_specific_transform(self, input_x, sequence_length = None):
        """Most kernels generate random features. For the LinearKernel,
        we merely return the input, adding an additional column for
        an intercept if specified.
        The sequence length argument is accepted for consistency with
        baseclass but is not used by this kernel.
        """
        if self.fit_intercept:
            if self.device == "cuda":
                xtrans = cp.zeros((input_x.shape[0], input_x.shape[1] + 1), cp.float64)
            else:
                xtrans = np.zeros((input_x.shape[0], input_x.shape[1] + 1), np.float64)
            xtrans[:,1:] = input_x
            return xtrans
        return input_x


    def kernel_specific_gradient(self, input_x, sequence_length = None):
        """The gradient for kernel-specific hyperparameters is calculated
        using an array (dz_dsigma) specific to each kernel.
        This kernel has no kernel-specific hyperparameters and hence
        can return a shape[1] == 0 array for gradient.
        """
        xtrans = self.transform_x(input_x)
        if self.device == "cuda":
            return xtrans, cp.zeros((xtrans.shape[0], 0, 0))
        return xtrans, np.zeros((xtrans.shape[0], 0, 0))



"""A quick and dirty timeit script for checking the time to perform graph
and polynomial feature generation."""
import math
import timeit
import numpy as np
import cupy as cp
import cupyx
from scipy.stats import chi
from cpu_rf_gen_module import cpuConv1dFGen
from cuda_rf_gen_module import gpuConv1dFGen


#Running the CPU numpy test with too many columns is extremely
#expensive -- avoid running that test by accident.
MAX_COLS_MATMUL = 124


def timetest():
    """Run a series of timeit tests and print the results."""
    #Change these lines depending on what dims you want to test.
    nthreads = 2
    nrows, ncols, nnodes, conv_width = 256, 100, 70, 9

    ntests = 100
    print("**************************************")
    print(f"Running test with {nrows} rows, {ncols} cols, {nnodes} nodes.")
    print("**************************************")


    cpu_graph_setup = f"""
nthreads = {nthreads}
import math
import numpy as np
from scipy.stats import chi
from cpu_rf_gen_module import cpuConv1dFGen
random_seed = 123
rng = np.random.default_rng(random_seed)
marr = rng.uniform(low=-10.0, high=10.0, size=({nrows},{nnodes},{ncols}))
marr = marr.astype(np.float32)
radem_array = np.asarray([-1.0, 1.0])
padded_width = 2**math.ceil(np.log2(max({ncols * conv_width}, 2)))
D1 = rng.choice(radem_array, size=(3,1,padded_width), replace=True)
D1 = D1.astype(np.int8)
chi_arr = chi.rvs(df={ncols}, size=padded_width, random_state=random_seed)
chi_arr = chi_arr.astype(np.float32)
from __main__ import graphconv_cpu_test"""
    print(f"Time(us) for GRAPH cpu with {nthreads} threads:")
    time_taken = timeit.timeit(f"graphconv_cpu_test(marr, D1, chi_arr, nthreads, {conv_width})",
            setup=cpu_graph_setup, number=ntests)
    print(1e6 * time_taken / ntests)


    random_seed = 123
    rng = np.random.default_rng(random_seed)
    marr = rng.uniform(low=-10.0, high=10.0, size=(nrows,nnodes,ncols))
    marr = marr.astype(np.float32)
    radem_array = np.asarray([-1.0, 1.0])
    padded_width = 2**math.ceil(np.log2(max(ncols * conv_width, 2)))
    D1 = rng.choice(radem_array, size=(3,1,padded_width), replace=True)
    D1 = D1.astype(np.int8)
    chi_arr = chi.rvs(df=ncols, size=padded_width, random_state=random_seed)
    chi_arr = chi_arr.astype(np.float32)
    marr, D1, chi_arr = cp.asarray(marr), cp.asarray(D1), cp.asarray(chi_arr)
    seqlen = cp.full(marr.shape[0], marr.shape[1]).astype(np.int32)
    print(f"Time(us) for GRAPH gpu with {nthreads} threads:")
    time_taken = cupyx.time.repeat(graphconv_gpu_test,
            (marr, D1, chi_arr, conv_width, seqlen),
            n_repeat = ntests)
    print(time_taken)


    
    cpu_poly_setup = f"""
nthreads = {nthreads}
import math
import numpy as np
from scipy.stats import chi
from cpu_convolution_float_hadamard_operations import floatCpuGraphPolyFHT
random_seed = 123
rng = np.random.default_rng(random_seed)
marr = rng.uniform(low=-10.0, high=10.0, size=({nrows},{nnodes},{ncols}))
marr = marr.astype(np.float32)
radem_array = np.asarray([-1.0, 1.0])
D1 = rng.choice(radem_array, size=(6,{1},{ncols}), replace=True)
D1 = D1.astype(np.int8)
chi_arr = chi.rvs(df={ncols}, size=(2,{ncols}), random_state=random_seed)
chi_arr = chi_arr.astype(np.float32)
from __main__ import poly_cpu_test"""
    #print(f"Time(us) for POLY cpu with {nthreads} threads:")
    #time_taken = timeit.timeit("poly_cpu_test(marr, D1, chi_arr, nthreads)",
    #        setup=cpu_poly_setup, number=ntests)
    #print(1e6 * time_taken / ntests)


    random_seed = 123
    rng = np.random.default_rng(random_seed)
    marr = rng.uniform(low=-10.0, high=10.0, size=(nrows,nnodes,ncols))
    marr = marr.astype(np.float32)
    radem_array = np.asarray([-1.0, 1.0])
    D1 = rng.choice(radem_array, size=(6,1,ncols), replace=True)
    D1 = D1.astype(np.int8)
    chi_arr = chi.rvs(df=ncols, size=(2,ncols), random_state=random_seed)
    chi_arr = chi_arr.astype(np.float32)
    marr, D1, chi_arr = cp.asarray(marr), cp.asarray(D1), cp.asarray(chi_arr)
    #print(f"Time(us) for POLY gpu with {nthreads} threads:")
    #time_taken = cupyx.time.repeat(poly_gpu_test, (marr, D1, chi_arr), n_repeat = ntests)
    #print(time_taken)



def graphconv_cpu_test(marr, diag, chi_arr, nthreads, conv_width):
    """Run the CPU graph convolution feature generation routine."""
    output_arr = np.zeros((marr.shape[0], diag.shape[2] * 2))
    seqlen = np.full(marr.shape[0], marr.shape[1]).astype(np.int32)
    cpuConv1dFGen(marr, seqlen, diag, output_arr, chi_arr, conv_width,
            nthreads, "none")


def graphconv_gpu_test(marr, diag, chi_arr, conv_width, seqlen):
    """Run the CPU graph convolution feature generation routine."""
    output_arr = cp.zeros((marr.shape[0], diag.shape[2] * 2))
    gpuConv1dFGen(marr, seqlen, diag, output_arr, chi_arr, conv_width,
            1, "none")

def poly_cpu_test(marr, diag, chi_arr, nthreads):
    output_arr = np.zeros((marr.shape[0], diag.shape[2]), dtype=np.float32)
    cpuGraphPolyFHT(marr, diag, chi_arr, output_arr, 2, nthreads)


def poly_gpu_test(marr, diag, chi_arr):
    output_arr = cp.zeros((marr.shape[0], diag.shape[2]), dtype=np.float32)
    gpuGraphPolyFHT(marr, diag, chi_arr, output_arr, 2, 2)


if __name__ == "__main__":
    timetest()



"""Implements a randomized nystrom preconditioner with
some methods that are useful for calculating NMLL
during tuning."""
try:
    import cupy as cp
except:
    pass
import numpy as np

from .rand_nys_constructors import initialize_srht, initialize_srht_multipass



class RandNysTuningPreconditioner():
    """Implements a preconditioner based on the randomized
    Nystrom approximation to the inverse of Z^T Z + lambda
    with some methods that are useful when tuning.
    """
    def __init__(self, kernel, dataset, max_rank, verbose,
                random_state = 123, method = "srht"):
        """Class constructor.

        Args:
            kernel: A valid kernel object that can generate random features.
            dataset: A valid online or offline dataset object that can retrieve
                chunked data.
            max_rank (int): The rank of the preconditioner.
            verbose (bool): If True, print regular updates.
            random_state (int): The random number seed for the random number
                generator.
            method (str): One of "srht", "srht_2". Determines the method of
                preconditioner construction.
        """
        self.device = kernel.device


        if method not in ["srht_2", "srht_3", "srht"]:
            raise ValueError("Unknown method supplied for tuning preconditioner "
                    "construction.")

        if method.startswith("srht_"):
            n_passes = int(method.split("_")[1])
            self.u_mat, self.eig, self.z_trans_y, self.y_trans_y = \
                            initialize_srht_multipass(dataset,
                                max_rank, kernel, random_state, verbose,
                                n_passes, get_zty = True)
        elif method == "srht":
            self.u_mat, self.eig, self.z_trans_y, self.y_trans_y = \
                            initialize_srht(dataset, max_rank,
                                kernel, random_state, verbose, get_zty = True)

        lambda_ = kernel.get_lambda()
        min_eig = self.eig.min()
        self.eig += lambda_**2

        self.inv_eig = self.eig.copy()
        mask = self.inv_eig > 1e-14
        self.inv_eig[mask] = 1 / self.inv_eig[mask]
        self.inv_eig[mask==False] = 0.0

        self.achieved_ratio = min_eig / lambda_**2
        self.prefactor = float(min_eig + lambda_**2)


    def batch_matvec(self, xvec):
        """Returns a matvec of the preconditioner with a set of input
        vectors xvec."""
        xprod = self.u_mat.T @ xvec
        xprod1 = self.u_mat @ (self.inv_eig[:,None] * self.prefactor * xprod)
        xprod2 = xvec - (self.u_mat @ xprod)
        return xprod2 + xprod1


    def get_rank(self):
        """Returns the preconditioner rank."""
        return self.inv_eig.shape[0]


    def get_logdet(self):
        """Returns the log determinant of the preconditioner. Useful
        for some hyperparameter tuning schemes."""
        logdet = 1 + (self.eig - self.prefactor) / self.prefactor
        if self.device == "cpu":
            return float(np.log(logdet.clip(min=1e-12)).sum())
        return float(cp.log(logdet.clip(min=1e-12)).sum())


    def matvec_for_sampling(self, xvec):
        """Returns self.u_mat @ eigvals[:,None] * x, where x is of shape
        (self.eig.shape[0], nsamples) and eigvals is the square root of
        the eigenvalues of the first term of the preconditioner."""
        eigvals = self.eig.copy()
        eigvals[eigvals < 0] = 0
        if self.device == "cpu":
            eigvals[:] = np.sqrt(eigvals)
        else:
            eigvals[:] = cp.sqrt(eigvals)
        prefactor = np.sqrt(1 / self.prefactor)
        xprod = self.u_mat.T @ xvec
        xprod1 = self.u_mat @ (eigvals[:,None] * prefactor * xprod)
        xprod2 = xvec - (self.u_mat @ xprod)
        return xprod1 + xprod2

    def get_zty(self):
        """Returns the product Z^T y, which is assembled during
        preconditioner construction."""
        return self.z_trans_y


    def get_yty(self):
        """Returns the product y^T y, which is assembled during
        preconditioner construction."""
        return float(self.y_trans_y)



"""Describes the AuxiliaryBaseclass from which other model classes inherit.

The AuxiliaryBaseclass describes class attributes and methods shared by
auxiliary tools like kernel_xPCA and kernel_FGen.
"""
import sys
import numpy as np

try:
    import cupy as cp
except:
    pass

from .kernels import KERNEL_NAME_TO_CLASS, ARR_3D_KERNELS
from .constants import constants



class AuxiliaryBaseclass():
    """The base class for auxiliary toolkits in xGPR, primarily
    for performing kernel PCA and kernel KMeans.

    Attributes:
        kernel: The kernel object for the posterior predictive mean. The class of
            this object will depend on the kernel specified by the user.
        device (str): One of "cuda", "cpu". The user can update this as desired.
            All predict / tune / fit operations are carried out using the
            current device.
        double_precision_fht (bool): Indicates whether we are using double precision.
        verbose (bool): If True, regular updates are printed during
            hyperparameter tuning and fitting.
    """

    def __init__(self, num_rffs:int, hyperparams, num_features:int,
                    kernel_choice:str = "RBF", device:str = "cpu",
                    kernel_settings:dict = constants.DEFAULT_KERNEL_SPEC_PARMS,
                    random_seed:int = 123, verbose:bool = True,
                    num_threads:int = 2,
                    double_precision_fht:bool = False):
        """Constructor.

        Args:
            num_rffs (int): The number of random Fourier features
                to use for the auxiliary device.
            hyperparams (np.ndarray): A numpy array containing the kernel-specific
                hyperparameter. If you have fitted an xGPR model, the first
                hyperparameter is in general not kernel specific, so
                my_model.get_hyperparams()[1:] will retrieve the hyperparameters you
                need. For most kernels there is only one kernel-specific hyperparameter.
                For kernels with no kernel-specific hyperparameter (e.g. arc-cosine
                and polynomial kernels), this argument is ignored.
            num_features (int): The number of features (i.e. the expected length
                of the last dimension) of typical input.
            dataset: A valid dataset object.
            kernel_choice (str): The kernel that the model will use.
                Defaults to 'RBF'. Must be in kernels.kernel_list.
                KERNEL_NAME_TO_CLASS.
            device (str): Determines whether calculations are performed on
                'cpu' or 'cuda'. The initial entry can be changed later
                (i.e. model can be transferred to a different device).
                Defaults to 'cpu'.
            kernel_settings (dict): Contains kernel-specific parameters --
                e.g. 'matern_nu' for the nu for the Matern kernel, or 'conv_width'
                for the conv1d kernel.
            random_seed (int): A seed for the random number generator.
            verbose (bool): If True, regular updates are printed
                during fitting and tuning. Defaults to True.
            num_threads (int): The number of threads to use for random feature generation
                if running on CPU. If running on GPU, this argument is ignored.
            double_precision_fht (bool): If True, use double precision during FHT for
                generating random features. For most problems, it is not beneficial
                to set this to True -- it merely increases computational expense
                with negligible benefit -- but this option is useful for testing.
                Defaults to False.
        """
        #We should never use a y-intercept for kPCA or clustering.
        kernel_settings["intercept"] = False

        self.verbose = verbose

        if kernel_choice not in KERNEL_NAME_TO_CLASS:
            raise ValueError("An unrecognized kernel choice was supplied.")

        if kernel_choice in ARR_3D_KERNELS:
            if "conv_width" in kernel_settings:
                xdim = (1, kernel_settings["conv_width"], num_features)
            else:
                xdim = (1, 10, num_features)
        else:
            xdim = (1, num_features)

        self.kernel = KERNEL_NAME_TO_CLASS[kernel_choice](xdim,
                            num_rffs, random_seed, device,
                            num_threads, double_precision_fht,
                            kernel_spec_parms = kernel_settings)

        self.double_precision_fht = double_precision_fht
        self.device = device
        full_hparams = self.kernel.get_hyperparams()
        if full_hparams.shape[0] > 1:
            full_hparams[1:] = hyperparams
        self.kernel.set_hyperparams(full_hparams)


    def pre_prediction_checks(self, input_x, sequence_lengths):
        """Checks input data to ensure validity.

        Args:
            input_x (np.ndarray): A numpy array containing the input data.
            sequence_lengths: None if you are using a fixed-vector kernel (e.g.
                RBF) and a 1d array of the number of elements in each sequence /
                nodes in each graph if you are using a graph or Conv1d kernel.

        Returns:
            x_array: A cupy array (if self.device is cuda) or a reference
                to the unmodified input array otherwise.

        Raises:
            ValueError: If invalid inputs are supplied,
                a detailed ValueError is raised to explain.
        """
        if not self.kernel.validate_new_datapoints(input_x):
            raise ValueError("The input has incorrect dimensionality.")
        if sequence_lengths is None:
            if len(input_x.shape) != 2:
                raise ValueError("sequence_lengths is required if using a "
                        "convolution kernel.")
        else:
            if len(input_x.shape) == 2:
                raise ValueError("sequence_lengths must be None if using a "
                    "fixed vector kernel.")

        if self.device == "cuda":
            mempool = cp.get_default_memory_pool()
            mempool.free_all_blocks()


    ####The remaining functions are all getters / setters.


    @property
    def device(self):
        """Property definition for the device attribute."""
        return self._device

    @device.setter
    def device(self, value):
        """Setter for the device attribute."""
        if value not in ["cpu", "cuda"]:
            raise ValueError("Device must be in ['cpu', 'cuda'].")

        if "cupy" not in sys.modules and value == "cuda":
            raise ValueError("You have specified the cuda fit mode but CuPy is "
                "not installed. Currently CPU only fitting is available.")

        if "xGPR.xgpr_cuda_rfgen_cpp_ext" not in sys.modules and value == "cuda":
            raise ValueError("You have specified the cuda fit mode but the "
                "cudaHadamardTransform module is not installed / "
                "does not appear to have installed correctly. "
                "Currently CPU only fitting is available.")

        self.kernel.device = value
        if value == "cuda":
            mempool = cp.get_default_memory_pool()
            mempool.free_all_blocks()
        self._device = value



"""Checks exact gradients against numerical gradients for
all currently available kernels."""
import unittest
#All currently available kernels are listed as keys in this dict.
from xGPR.kernels import KERNEL_NAME_TO_CLASS
from .kernel_specific_gradient_test import run_kernelspecific_test

class CheckKernelGradients(unittest.TestCase):
    """Checks the NMLL gradients for all currently implemented
    kernels."""

    def test_kernel_gradients(self):
        for kernel_name in KERNEL_NAME_TO_CLASS.keys():
            is_conv_kernel = "conv" in kernel_name.lower() or "graph" \
                    in kernel_name.lower()
            costcomps = run_kernelspecific_test(kernel_name,
                        conv_kernel = is_conv_kernel)
            for costcomp in costcomps:
                self.assertTrue(costcomp)

            #For convolution kernels, also test that graph or sequence averaging
            #works.
            if is_conv_kernel and "TwoLayer" not in kernel_name:
                print("****Testing with averaging****")
                costcomps = run_kernelspecific_test(kernel_name,
                        conv_kernel = is_conv_kernel,
                        averaging = 'full')
                for costcomp in costcomps:
                    self.assertTrue(costcomp)



if __name__ == "__main__":
    unittest.main()



"""These are convolution kernels that can be run once on the input, and
the resulting features can be saved and then fed into another kernel,
resulting in a two-layer GP. Consequently, these can also be considered
'feature extractors', hence their names."""
from math import ceil

import numpy as np
from scipy.stats import chi

from xGPR.xgpr_cpu_rfgen_cpp_ext import cpuConv1dMaxpool
try:
    import cupy as cp
    from xGPR.xgpr_cuda_rfgen_cpp_ext import cudaConv1dMaxpool
except:
    pass



class FHTMaxpoolConv1dFeatureExtractor():
    """A ReLU activation, global maxpool kernel that can be run
    once on the data (since it does not need any separate
    hyperparameters). It should never be used as a model kernel,
    only as a feature extractor; it is missing some attributes
    that would be required for a model kernel (e.g. gradient
    calcs).

    Attributes:
        conv_width (int): The width of the convolution kernel.
            This hyperparameter can be set based on experimentation
            or domain knowledge.
        radem_diag: The diagonal matrices for the SORF transform. Type is int8.
        chi_arr: A diagonal array whose elements are drawn from the chi
            distribution. Ensures the marginals of the matrix resulting
            from S H D1 H D2 H D3 are correct.
        num_threads (int): Number of threads to use if running on CPU;
            ignored if running on GPU.
    """

    def __init__(self, seqwidth, num_rffs, random_seed = 123, device = "cpu",
                    conv_width = 9, num_threads = 2):
        """Constructor for FHT_Conv1d.

        Args:
            seqwidth (int): The number of features per timepoint / sequence element.
            num_rffs (int): The user-requested number of random Fourier features.
                For sine-cosine kernels (RBF, Matern), this will be saved by the
                class as num_rffs.
            random_seed (int): The seed to the random number generator.
            device (str): One of 'cpu', 'cuda'. Indicates the starting device.
            conv_width (int): The width of the convolution kernel. Defaults to 9.
            num_threads (int): Number of threads to use if running on CPU;
                ignored if running on GPU.

        Raises:
            ValueError: A ValueError is raised if the dimensions of the input are
                inappropriate given the conv_width.
        """
        rng = np.random.default_rng(random_seed)
        self.conv_width = conv_width

        self.num_rffs = num_rffs
        self.seqwidth = seqwidth
        dim2_no_padding = conv_width * seqwidth
        padded_dims = 2**ceil(np.log2(max(dim2_no_padding, 2)))
        init_calc_featsize = ceil(self.num_rffs / padded_dims) * \
                        padded_dims

        radem_array = np.asarray([-1,1], dtype=np.int8)

        self.radem_diag = rng.choice(radem_array, size=(3, 1, init_calc_featsize),
                                replace=True)
        self.chi_arr = chi.rvs(df=padded_dims, size=self.num_rffs,
                            random_state = random_seed).astype(np.float32)

        self.num_threads = num_threads
        self.device = device


    def kernel_specific_set_device(self, new_device):
        """Called by parent class when device is changed. Moves
        some of the object parameters to the appropriate device
        and updates convenience references to numpy / cupy
        functions used for generating features."""
        if new_device == "cuda":
            self.radem_diag = cp.asarray(self.radem_diag)
            self.chi_arr = cp.asarray(self.chi_arr)
        elif not isinstance(self.radem_diag, np.ndarray):
            self.radem_diag = cp.asnumpy(self.radem_diag)
            self.chi_arr = cp.asnumpy(self.chi_arr)


    def transform_x(self, input_x, sequence_length):
        """Performs the feature generation."""
        if sequence_length is None:
            raise ValueError("sequence_length is required for convolution kernels.")
        if input_x.shape[2] != self.seqwidth:
            raise ValueError("Unexpected number of features per timepoint / sequence element "
                    "on this input.")

        sequence_length = sequence_length.astype(np.int32, copy=False)

        if self.device == "cpu":
            output_x = np.zeros((input_x.shape[0], self.num_rffs), np.float32)
            x_in = np.ascontiguousarray(input_x.astype(np.float32, copy=False))
            cpuConv1dMaxpool(x_in, output_x, self.radem_diag, self.chi_arr,
                    sequence_length, self.conv_width, self.num_threads)
        else:
            output_x = cp.zeros((input_x.shape[0], self.num_rffs), cp.float32)
            x_in = cp.ascontiguousarray(cp.asarray(input_x).astype(cp.float32, copy=False))
            cudaConv1dMaxpool(x_in, output_x, self.radem_diag, self.chi_arr,
                    sequence_length, self.conv_width)

        return output_x


    @property
    def device(self):
        """Getter for the device property, which determines
        whether calculations are on CPU or GPU."""
        return self.device_

    @device.setter
    def device(self, value):
        """Setter for device, which determines whether calculations
        are on CPU or GPU.

        Args:
            value (str): Must be one of 'cpu', 'cuda'.

        Raises:
            ValueError: A ValueError is raised if an unrecognized
                device is passed.
        """
        if value not in ('cuda', 'cpu'):
            raise ValueError("Unrecognized device supplied. Must be one "
                    "of 'cpu', 'cuda'.")
        self.device_ = value
        self.kernel_specific_set_device(value)



"""Provides data for classification-specific unit tests. We test
all the kernels in regression-specific unit tests, so for classification,
we use a fixed-vector dataset only with the RBF kernel. The wine
dataset is used here as a (horrifically) simple test."""
import sklearn
from sklearn import datasets
from sklearn.preprocessing import StandardScaler
import numpy as np

from xGPR.data_handling.dataset_builder import build_classification_dataset

RANDOM_STATE = 123


def build_discriminant_traintest_split():
    """Loads the wine dataset from sklearn and returns a train-test
    split of the data.

    Returns:
        online_data (OnlineDataset): The raw data stored in memory.
    """
    xvalues, yvalues = sklearn.datasets.load_wine(return_X_y = True)
    scaler = StandardScaler()

    #Scaling train and test data together is ordinarily not great, but
    #this is just for testing purposes.
    xvalues = scaler.fit_transform(xvalues)

    rng = np.random.default_rng(123)
    idx = rng.permutation(xvalues.shape[0])
    xvalues, yvalues = xvalues[idx,:], yvalues[idx]
    cutoff = int(0.75 * idx.shape[0])

    if len(xvalues.shape) == 2:
        sequence_lengths = None
        train_data = build_classification_dataset(xvalues[:cutoff,...],
                yvalues[:cutoff], chunk_size = 2000)
        test_data = build_classification_dataset(xvalues[cutoff:,...],
                yvalues[cutoff:], chunk_size = 2000)

    else:
        sequence_lengths = np.full(xvalues.shape[0], xvalues.shape[1]).astype(np.int32)
        train_data = build_classification_dataset(xvalues[:cutoff,...],
                yvalues[:cutoff], sequence_lengths[:cutoff], chunk_size = 2000)
        test_data = build_classification_dataset(xvalues[cutoff:,...],
                yvalues[cutoff:], sequence_lengths[cutoff:], chunk_size = 2000)


    return train_data, test_data



"""Tests the static layer construction to ensure they
initialize and can transform data."""
import sys
import unittest
import os

from xGPR.static_layers import FastConv1d

sys.path.append(os.path.join(os.path.dirname(os.path.abspath(__file__)), ".."))
from utils.build_test_dataset import build_test_dataset


RANDOM_SEED = 123


class CheckStatLayerConstruction(unittest.TestCase):
    """Tests construction of static layer objects."""

    def test_static_layer_builders(self):
        """Test static layer construction and basic functions."""
        test_online_dataset, _ = build_test_dataset(conv_kernel = True,
            xsuffix = "testxvalues.npy", ysuffix = "testyvalues.npy")
        train_online_dataset, _ = build_test_dataset(conv_kernel = True)

        conv_statlayer = FastConv1d(test_online_dataset.get_xdim()[2],
                device = "cpu", random_seed = RANDOM_SEED, conv_width = 3,
                num_features = 512)

        xchunks = list(train_online_dataset.get_chunked_x_data())
        x_trans = conv_statlayer.predict(xchunks[0][0], xchunks[0][1])
        self.assertTrue(x_trans.shape[1] == 512)
        self.assertTrue(xchunks[0][0].shape[0] == x_trans.shape[0])


if __name__ == "__main__":
    unittest.main()



"""Evaluates a tuned, fitted model on an existing
dataset where expected performance is known."""
from scipy.stats import spearmanr



def evaluate_model(model, train_dataset, test_dataset, get_var = True):
    """Check how well model predictions align with gt, using
    an existing dataset where expected performance is 'known'."""
    if not get_var:
        preds = model.predict(test_dataset._xdata,
                test_dataset._sequence_lengths, get_var)
    else:
        preds, _ = model.predict(test_dataset._xdata,
                test_dataset._sequence_lengths, get_var)
    y_test = test_dataset._ydata * train_dataset._trainy_std
    y_test += train_dataset._trainy_mean
    return spearmanr(preds, y_test)[0]



"""Checks the fast Hadamard transform based routine
used by static layers."""
import sys

import unittest
import numpy as np

from xGPR.xgpr_cpu_rfgen_cpp_ext import cpuConv1dMaxpool
from xGPR.xgpr_cuda_rfgen_cpp_ext import cudaConv1dMaxpool
import cupy as cp

from conv_testing_functions import get_features_maxpool, get_initial_matrices_fht



class TestMaxpoolFeatureGen(unittest.TestCase):
    """Tests feature generation for static layers."""


    def test_conv1d_maxpool(self):
        """Tests the C / Cuda FHT-based convolution with global max pooling
        functions."""
        kernel_width, num_aas, aa_dim, num_freqs = 9, 23, 21, 130
        sigma, ndatapoints = 1, 120
        outcomes = run_maxpool_evaluation(ndatapoints, kernel_width, aa_dim, num_aas,
                    num_freqs, sigma, mode = "maxpool")
        for outcome in outcomes:
            self.assertTrue(outcome)

        kernel_width, num_aas, aa_dim, num_freqs = 15, 23, 1060, 8194
        sigma, ndatapoints = 1, 5
        outcomes = run_maxpool_evaluation(ndatapoints, kernel_width, aa_dim, num_aas,
                    num_freqs, sigma, mode = "maxpool")
        for outcome in outcomes:
            self.assertTrue(outcome)

        kernel_width, num_aas, aa_dim, num_freqs = 5, 56, 2, 62
        sigma, ndatapoints = 1, 100

        outcomes = run_maxpool_evaluation(ndatapoints, kernel_width, aa_dim, num_aas,
                    num_freqs, sigma, mode = "maxpool", precision = "float")
        for outcome in outcomes:
            self.assertTrue(outcome)


        kernel_width, num_aas, aa_dim, num_freqs = 5, 56, 256, 500
        sigma, ndatapoints = 1, 232

        outcomes = run_maxpool_evaluation(ndatapoints, kernel_width, aa_dim, num_aas,
                    num_freqs, sigma, mode = "maxpool")
        for outcome in outcomes:
            self.assertTrue(outcome)

        kernel_width, num_aas, aa_dim, num_freqs = 5, 512, 1024, 1024
        sigma, ndatapoints = 1, 11

        outcomes = run_maxpool_evaluation(ndatapoints, kernel_width, aa_dim, num_aas,
                    num_freqs, sigma, mode = "maxpool")
        for outcome in outcomes:
            self.assertTrue(outcome)


def run_maxpool_evaluation(ndatapoints, kernel_width, aa_dim, num_aas,
                    num_freqs, sigma, mode, precision = "double"):
    """Run an evaluation for the ReLU / maxpool feature generation
    routine, primarily used for static layers."""
    dim2, num_blocks, xdata, seqlen, features, s_mat, \
                radem = get_initial_matrices_fht(ndatapoints, kernel_width,
                        aa_dim, num_aas, num_freqs, mode, precision)
    true_features = get_features_maxpool(xdata, kernel_width, dim2,
                            radem, s_mat, num_freqs, num_blocks,
                            seqlen, precision)

    features = features.astype(np.float32)
    cpuConv1dMaxpool(xdata, features, radem, s_mat, seqlen,
                kernel_width, 2)

    outcome = check_results(true_features, features, precision)
    print(f"Settings: N {ndatapoints}, kernel_width {kernel_width}, "
        f"aa_dim: {aa_dim}, num_aas: {num_aas}, num_freqs: {num_freqs}, "
        f"sigma: {sigma}, mode: {mode}, precision {precision}\n"
        f"Does result match on CPU? {outcome}")

    if "cupy" not in sys.modules:
        return [outcome]

    xdata = cp.asarray(xdata)
    features[:] = 0
    features = cp.asarray(features)
    s_mat = cp.asarray(s_mat)
    radem = cp.asarray(radem)

    cudaConv1dMaxpool(xdata, features, radem, s_mat, seqlen,
                kernel_width)
    outcome_cuda = check_results(true_features, features, precision)
    print(f"Settings: N {ndatapoints}, kernel_width {kernel_width}, "
        f"aa_dim: {aa_dim}, num_aas: {num_aas}, num_freqs: {num_freqs}, "
        f"sigma: {sigma}, mode: {mode}, precision {precision}\n"
        f"Does result match on cuda? {outcome_cuda}")

    print("\n")
    return outcome, outcome_cuda


def check_results(gt_array, test_array, precision):
    """Checks a ground truth array against a test array. We have
    to use different tolerances for 32-bit vs 64 since 32-bit
    can vary slightly."""
    if precision == "float":
        return np.allclose(gt_array, test_array, rtol=1e-6, atol=1e-6)
    return np.allclose(gt_array, test_array)



if __name__ == "__main__":
    unittest.main()



"""Provides the FastConv1d class for convolution-based feature extraction.

If a GP is equivalent to an infinitely wide 2-layer NN, we can turn it into
an infinitely wide 3-layer NN by using feature extraction then feeding
the result into the GP. The convolution-based feature extractors here
are designed to work on time series and sequences. They act like a convolutional
layer in an NN and generate a fixed-length vector so that the GP then functions
like a fully connected layer on top of a convolutional layer.
"""
import sys

import numpy as np
try:
    import cupy as cp
except:
    pass

from ..kernels.convolution_kernels.conv_feature_extractor import FHTMaxpoolConv1dFeatureExtractor


class FastConv1d:
    """Provides tools for performing convolution-based feature
    extraction on an input dataset. The input dataset should
    have x-values that are 3d arrays where shape[1] is timepoints
    or sequence locations and shape[2] is features.

    Attributes:
        conv_kernel: Kernel object of appropriate class
            (depending on the kernel type selected).
        device (str): One of ['cpu', 'cuda']. Indicates the current device
            for the FastConv1d.
        seq_width (int): The anticipated width (number of features) of
                each input sequence / series.
        num_features (int): The number of random features to generate.
            More = improved performance but slower feature extraction
            and slower model training.
    """

    def __init__(self, seq_width:int, device:str = "cpu", random_seed:int = 123,
            conv_width:int = 9, num_features:int = 512):
        """Constructor for the FastConv1d class.

        Args:
            seq_width (int): The anticipated width (number of features) of
                each input sequence / series.
            device (str): Must be one of ['cpu', 'cuda']. Indicates the current
                device for the feature extractor.
            random_seed (int): The seed to the random number generator.
                Defaults to 123.
            conv_width (list): A convolution kernel width.
            num_features (int): The number of random features to generate.
                More = improved performance but slower feature extraction
                and slower model training.

        Raises:
            ValueError: If an unrecognized kernel type or other invalid
                input is supplied.
        """
        self.seq_width = seq_width
        self.num_features = num_features

        self.conv_kernel = FHTMaxpoolConv1dFeatureExtractor(seq_width,
                            self.num_features, random_seed, device = device,
                            conv_width = conv_width)
        self.device = device


    def predict(self, x_array, sequence_lengths, chunk_size:int = 2000):
        """Performs feature extraction using a 1d convolution kernel
        and returns an array containing the result. This function should
        be used if it is desired to generate features for sequence /
        timeseries data AFTER training (i.e. when making predictions).
        Note that when training, you should use conv1d_pretrain_feat_extract,
        which takes a dataset as input rather than an array.

        Args:
            x_array: A numpy array. Should be a 3d array with same shape[1]
                and shape[2] as the training set.
            sequence_lengths (np.ndarray): A 1d numpy array with shape[0] ==
                shape[0] of x_array. Indicates the number of sequence elements
                in each datapoint of x_array (so that x_array can be zero-padded).
            chunk_size (int): The batch size in which the input data array
                will be processed. This limits memory consumption. Defaults
                to 2000.

        Returns:
            x_features: A 2d numpy array of shape (N, M) for
                N datapoints, M features that results from applying
                the feature extraction operation to the input.

        Raises:
            ValueError: If the inputs are not valid a detailed ValueError
                is raised explaining the issue.
        """
        if sequence_lengths.shape[0] != x_array.shape[0]:
            raise ValueError("The shape[0] of sequence_lengths must match the shape[0] "
                    "of x_array.")

        x_features = []
        for i in range(0, x_array.shape[0], chunk_size):
            cutoff = min(x_array.shape[0], i + chunk_size)
            if cutoff - i == 0:
                continue

            xtrans = self.conv_kernel.transform_x(x_array[i:cutoff,...],
                    sequence_lengths[i:cutoff])

            if self.device == "cuda":
                xtrans = cp.asnumpy(xtrans)

            x_features.append(xtrans)

        if self.device == "cuda":
            mempool = cp.get_default_memory_pool()
            mempool.free_all_blocks()

        x_features = np.vstack(x_features)
        return x_features


    @property
    def device(self):
        """Property definition for the device attribute."""
        return self.device_



    @device.setter
    def device(self, value):
        """Setter for the device attribute."""
        if value not in ["cpu", "cuda"]:
            raise ValueError("Device must be in ['cpu', 'cuda'].")

        if "cupy" not in sys.modules and value == "cuda":
            raise ValueError("You have specified the cuda fit mode but CuPy is "
                "not installed. Currently CPU only fitting is available.")

        self.conv_kernel.device = value
        self.device_ = value



"""This module 'compresses' the output of another kernel using
SRHT operations. This can be useful if we want to project from
a high-dimensional random feature space back into a lower
more manageable space for hyperparameter tuning purposes."""
from math import ceil

import numpy as np

from xGPR.xgpr_cpu_rfgen_cpp_ext import cpuRBFFeatureGen, cpuRBFGrad
from xGPR.xgpr_cpu_rfgen_cpp_ext import cpuSRHT
try:
    import cupy as cp
    from xGPR.xgpr_cuda_rfgen_cpp_ext import cudaSRHT
except:
    pass


class SRHTCompressor():
    """This class provides the tools needed to compress the output
    of another kernel to a specified dimensionality.

    Attributes:
        compression_size (int): The desired size for the output.
        input_size (int): The expected size of the input.
        padded_dims (int): The next largest power of two greater
            than input_size. Used for padding the input.
        radem: A numpy or cupy diagonal matrix drawn from a
            Rademacher distribution of shape (padded_dims).
        col_sampler (np.ndarray): A numpy array of shape (compression_size)
            that permutes the columns of the compressed data.
        compressor_func: A reference to an appropriate wrapped C++ function.
        device: Either "cpu" or "cuda".
        double_precision (bool): If True, input is assumed to be
                doubles, else floats. Right now set to True by default.
        num_threads (int): Max number of threads to use for CPU operations.
    """

    def __init__(self, compression_size, input_size, device = "cpu",
            double_precision = True, random_seed = 123, num_threads = 2):
        """Class constructor.

        Args:
            compression_size (int): The desired size for the output.
            input_size (int): The expected size of the input.
            device (str): The starting device.
            double_precision (bool): If True, input is assumed to be
                doubles, else floats.
            random_seed (int): A seed for the random number generator.
            num_threads (int): The max number of threads to use for CPU
                operations.

        Raises:
            ValueError: Raises a ValueError if the inputs are inappropriate.
        """
        self.compression_size = compression_size
        self.input_size = input_size
        if compression_size >= input_size or compression_size <= 1:
            raise ValueError("The compression size should be < "
                    "the number of rffs and > 1.")
        self.padded_dims = 2**ceil(np.log2(max(input_size, 2)))
        self.double_precision = double_precision

        radem_array = np.asarray([-1,1], dtype=np.int8)
        rng = np.random.default_rng(random_seed)
        self.radem = rng.choice(radem_array, size=(self.padded_dims),
                                replace=True)
        self.col_sampler = rng.permutation(self.padded_dims)
        self.truncated_sampler = self.col_sampler[:self.compression_size]
        self.compressor_func = None
        self.device = device
        self.num_threads = num_threads


    def transform_x(self, features, no_compression = False):
        """Performs the SRHT operation on the input and
        returns the downsized array.

        Args:
            features: A 2d cupy or numpy array (as appropriate for device)
                that will be compressed along the second dimension. Will
                be zero-padded if not already a power of 2.
            no_compression (bool): If True, do not truncate.

        Returns:
            features: A 2d cupy or numpy array containing the compressed
                feature information.
        """
        if features.shape[1] != self.input_size or len(features.shape) != 2:
            raise ValueError("Input with unexpected size passed to a compressor "
                    "module.")
        if features.shape[1] < self.padded_dims:
            xfeatures = self.zero_arr((features.shape[0], self.padded_dims), self.dtype)
            xfeatures[:,:features.shape[1]] = features
        else:
            xfeatures = features.astype(self.dtype)

        self.compressor_func(xfeatures, self.radem, self.num_threads)
        if no_compression:
            return xfeatures[:,self.col_sampler]

        return xfeatures[:,self.truncated_sampler]


    @property
    def device(self):
        """Getter for the device property, which determines
        whether calculations are on CPU or GPU."""
        return self.device_


    @device.setter
    def device(self, value):
        """Setter for device, which determines whether calculations
        are on CPU or GPU. Note that each kernel must also have
        a kernel_specific_set_device function (enforced via
        an abstractmethod) to make any kernel-specific changes
        that occur when the device is switched.

        Args:
            value (str): Must be one of 'cpu', 'cuda'.

        Raises:
            ValueError: A ValueError is raised if an unrecognized
                device is passed.

        Note that a number of 'convenience attributes' (e.g. self.dtype,
        self.zero_arr) are set as references to either cupy or numpy functions.
        This avoids having to write two sets of functions (one for cupy, one for
        numpy) when the steps involved are the same.
        """
        if value == "cpu":
            if not isinstance(self.radem, np.ndarray):
                self.radem = cp.asnumpy(self.radem)
            self.zero_arr = np.zeros
            self.compressor_func = cpuSRHT
            if self.double_precision:
                self.dtype = np.float64
            else:
                self.dtype = np.float32

        elif value == "cuda":
            self.radem = cp.asarray(self.radem)
            self.zero_arr = cp.zeros
            self.compressor_func = cudaSRHT
            if self.double_precision:
                self.dtype = cp.float64
            else:
                self.dtype = cp.float32
        else:
            raise ValueError("Unrecognized device supplied. Must be one "
                    "of 'cpu', 'cuda'.")
        self.device_ = value



"""Implements functions for fitting the weights using exact calculations,
which is generally only recommended for small datasets & numbers of
random features, and for fitting the variance using exact calculations."""
try:
    import cupy as cp
    import cupyx as cpx
except:
    pass
import numpy as np
from scipy.linalg import cho_solve

from ..scoring_toolkit.exact_nmll_calcs import calc_var_design_mat
from ..scoring_toolkit.exact_nmll_calcs import direct_weight_calc, calc_design_mat



def calc_weights_exact(dataset, kernel):
    """Calculates the weights when fitting the model using
    matrix decomposition. Exact and fast for small numbers
    of random features but poor scaling.

    Args:
        dataset: Either OnlineDataset or OfflineDataset,
            containing the information on the dataset we
            are fitting.
        kernel: A valid kernel object that can generate random
            features.

    Returns:
        weights: A cupy or numpy array of shape (M) for M
            random features.
    """
    z_trans_z, z_trans_y, _ = calc_design_mat(dataset, kernel)
    lambda_p = kernel.get_hyperparams(logspace=False)[0]
    z_trans_z.flat[::z_trans_z.shape[0]+1] += lambda_p**2
    _, weights = direct_weight_calc(z_trans_z, z_trans_y, kernel)
    return weights, 1, []



def calc_variance_exact(kernel, dataset, kernel_choice, variance_rffs):
    """Calculates the var matrix used for calculating
    posterior predictive variance on new datapoints. We
    only ever use closed-form matrix-decomposition based
    calculations here since the variance does not need to
    be approximated as accurately as the posterior predictive
    mean, so we can restrict the user to a smaller number of
    random features (defined in constants.constants).

    Args:
        kernel: A valid kernel object that can generate random features.
        dataset: Either an OnlineDataset or an OfflineDataset containing
            the data that needs to be fitted.
        kernel_choice (str): The type of kernel.
        variance_rffs (int): The number of random features for variance.

    Returns:
        var: A cupy or numpy array of shape (M, M) where M is the
            number of random features.
    """
    z_trans_z = calc_var_design_mat(dataset, kernel,
                        variance_rffs)
    lambda_ = kernel.get_lambda()
    z_trans_z.flat[::z_trans_z.shape[0]+1] += lambda_**2
    if kernel.device == "cpu":
        var = np.linalg.pinv(z_trans_z)
    else:
        var = cp.linalg.pinv(z_trans_z)
    return var


def calc_discriminant_weights_exact(dataset, kernel, x_mean,
        targets):
    """Calculates the weights when fitting a discriminant using
    exact matrix decomposition. Fast for small numbers of random
    features but poor scaling.

    Args:
        dataset: Either OnlineDataset or OfflineDataset,
            containing the information on the dataset we
            are fitting.
        kernel: A valid kernel object that can generate random
            features.
        x_mean (ndarray): An array of shape (num_rffs) containing
            the mean of the training data.
        targets (ndarray): A (num_rffs, nc) for nc classes shape
            array containing class specific means.


    Returns:
        weights: A cupy or numpy array of shape (M, nc) for M
            random features and nc classes.
    """
    num_rffs = kernel.get_num_rffs()
    if kernel.device == "cpu":
        z_trans_z = np.zeros((num_rffs, num_rffs))
    else:
        z_trans_z = cp.zeros((num_rffs, num_rffs))

    for i, (xdata, ldata) in enumerate(dataset.get_chunked_x_data()):
        xfeatures = kernel.transform_x(xdata, ldata) - x_mean[None,:]
        z_trans_z += xfeatures.T @ xfeatures
        if i % 2 == 0:
            if kernel.device == "cuda":
                mempool = cp.get_default_memory_pool()
                mempool.free_all_blocks()

    z_trans_z /= float(dataset.get_ndatapoints())
    z_trans_z.flat[::z_trans_z.shape[0]+1] += kernel.get_lambda()**2

    if kernel.device == "cpu":
        chol_z_trans_z = np.linalg.cholesky(z_trans_z)
        weights = cho_solve((chol_z_trans_z, True), targets)
    else:
        chol_z_trans_z = cp.linalg.cholesky(z_trans_z)
        weights = cpx.scipy.linalg.solve_triangular(chol_z_trans_z,
                        targets, lower=True)
        weights = cpx.scipy.linalg.solve_triangular(chol_z_trans_z.T,
                        weights, lower=False)
    return weights



"""This file lists key currently implemented
kernels, together with the performance (the spearman's r)
expected from each on the test set included with the
module. Add to this when a new kernel is added to
ensure that kernel is included; you will need to
decide what is 'acceptable performance', likely
based on an initial experiment. Note that kernels
which are essentially modified versions of RBF
kernels are excluded (e.g. Conv1dMatern does
not need to be tested since Conv1dRBF is)."""

#Each dictionary value contains 1) whether this
#is a convolution kernel and 2) the expected
#minimum performance.
IMPLEMENTED_KERNELS = {
        "Conv1dRBF":(True,0.58),"RBF":(False, 0.58),
        "Matern":(False, 0.55),
        "Linear":(False, 0.53),
        "RBFLinear":(False,0.55),
        "MiniARD":(False, 0.64),
        "Conv1dRBF":(True, 0.58),
        "GraphRBF":(True, 0.38),
        }



"""Tests the main tuning algorithm to ensure
we get performance on par with expectations."""
import sys
import os
import unittest

import numpy as np

from xGPR import xGPRegression as xGPReg
sys.path.append(os.path.join(os.path.dirname(os.path.abspath(__file__)), ".."))
from utils.build_test_dataset import build_test_dataset
from utils.model_constructor import get_models


class CheckTuning(unittest.TestCase):
    """Tests the tuning algorithm."""

    def test_tuning(self):
        """Test the tuning algorithm using an
        RBF kernel for simplicity (tuning & fitting with other
        kernels is tested under the complete pipeline tests."""
        online_data, _ = build_test_dataset(conv_kernel = False)
        models = [m for m in get_models("RBF", online_data) if m is not None]

        for mod in models:
            mod.verbose = False
            for tuning_method, nmll_method, max_iter in [("Nelder-Mead", "exact", 100),
                        ("Powell", "exact", 100),
                        ("L-BFGS-B", "exact", 100)]:
                _, niter, best_score = mod.tune_hyperparams(online_data,
                        tuning_method = tuning_method, n_restarts=1,
                        starting_hyperparams = np.array([0., 0.]),
                        max_iter = max_iter, nmll_method = nmll_method)

                print(mod.get_hyperparams())
                print("*************")
                print(f"{tuning_method}, {mod.device}, {nmll_method}")
                print(f"Best score: {best_score}")
                print(f"{niter} iterations\n\n\n")
                self.assertTrue(best_score < 430)


if __name__ == "__main__":
    unittest.main()



"""Provides a set of functions for computing exact NMLL,
exact NMLL gradients, and matrices and vectors useful
for these tasks. Some of these tools can also be used
for (exact) fitting routines."""
try:
    import cupy as cp
    import cupyx as cpx
except:
    pass
import numpy as np
from scipy.linalg import cho_solve

def calc_zty(dataset, kernel):
    """Calculates the vector Z^T y.

    Args:
        dataset: An Dataset object that can supply
            chunked data.
        kernel: A valid kernel object that can generate
            random features.
        device (str): One of "cpu", "cuda".

    Returns:
        z_trans_y (array): A shape (num_rffs)
            array that contains Z^T y.
        y_trans_y (float): The value y^T y.
    """
    if kernel.device == "cuda":
        z_trans_y = cp.zeros((kernel.get_num_rffs()))
    else:
        z_trans_y = np.zeros((kernel.get_num_rffs()))

    y_trans_y = 0

    for xin, yin, ldata in dataset.get_chunked_data():
        zdata, ydata = kernel.transform_x_y(xin, yin, ldata)
        z_trans_y += zdata.T @ ydata
        y_trans_y += float( (ydata**2).sum() )
    return z_trans_y, y_trans_y


def calc_design_mat(dataset, kernel):
    """Calculates the z_trans_z (z^T z) matrix where Z is the random
    features generated from raw input data X. Also generates
    y^T y and z^T y.

    Args:
        dataset: An OnlineDataset or OfflineDataset object storing
            the data we will use.
        kernel: A valid kernel object that can generate
            random features.

    Returns:
        z_trans_z: The cupy or numpy matrix resulting from z^T z. Will
            be shape M x M for M random features.
        z_trans_y: The cupy or numpy length M array resulting from
            z^T y for M random features.
        y_trans_y (float): The result of the dot product of y with itself.
            Used for some marginal likelihood calculations.
    """
    num_rffs = kernel.get_num_rffs()
    if kernel.device == "cpu":
        z_trans_z, z_trans_y = np.zeros((num_rffs, num_rffs)), np.zeros((num_rffs))
    else:
        z_trans_z = cp.zeros((num_rffs, num_rffs))
        z_trans_y = cp.zeros((num_rffs))
    y_trans_y = 0.0
    for i, (xin, yin, ldata) in enumerate(dataset.get_chunked_data()):
        xfeatures, ydata = kernel.transform_x_y(xin, yin, ldata)
        z_trans_y += xfeatures.T @ ydata
        z_trans_z += xfeatures.T @ xfeatures
        y_trans_y += ydata.T @ ydata
        if i % 2 == 0:
            if kernel.device == "cuda":
                mempool = cp.get_default_memory_pool()
                mempool.free_all_blocks()
                
    return z_trans_z, z_trans_y, float(y_trans_y)



def direct_weight_calc(chol_z_trans_z, z_trans_y, kernel):
    """Calculates the cholesky decomposition of (z^T z + lambda)^-1
    and then uses this to calculate the weights as (z^T z + lambda)^-1 z^T y.
    This exact calculation is only suitable for < 10,000 random features or so;
    cholesky has O(M^3) scaling.

    Args:
        z_trans_z: An M x M cupy or numpy matrix where M is the number of
            random features formed from z^T z when z is the random features
            generated for raw input data X.
        z_trans_y: A length M cupy or numpy array where M is the number of
            random features, formed from z^T y.
        kernel: A valid kernel object that can generate random features.

    Returns:
        chol_z_trans_z: The cholesky decomposition of z_trans_z. An M x M
            cupy or numpy matrix.
        weights: A length M cupy or numpy array containing the weights.
    """
    lambda_p = kernel.get_hyperparams(logspace=False)[0]
    chol_z_trans_z.flat[::chol_z_trans_z.shape[0]+1] += lambda_p**2
    if kernel.device == "cpu":
        chol_z_trans_z = np.linalg.cholesky(chol_z_trans_z)
        weights = cho_solve((chol_z_trans_z, True), z_trans_y)
    else:
        chol_z_trans_z = cp.linalg.cholesky(chol_z_trans_z)
        weights = cpx.scipy.linalg.solve_triangular(chol_z_trans_z,
                        z_trans_y, lower=True)
        weights = cpx.scipy.linalg.solve_triangular(chol_z_trans_z.T,
                        weights, lower=False)
    return chol_z_trans_z, weights



def calc_var_design_mat(dataset, kernel, variance_rffs):
    """Calculates the z_trans_z (z^T z) matrix where Z is the random
    features generated from raw input data X, for calculating
    variance only (since in this case we only use up to
    self.variance_rffs of the features generated).

    Args:
        dataset: An OnlineDataset or OfflineDataset object storing
            the data we will use.
        kernel: A valid kernel object that can generate random features.
        variance_rffs (int): The number of variance random features.

    Returns:
        z_trans_z: The cupy or numpy matrix resulting from z^T z. Will
            be shape M x M for M random features.
    """
    if kernel.device == "cpu":
        z_trans_z = np.zeros((variance_rffs, variance_rffs))
    else:
        z_trans_z = cp.zeros((variance_rffs, variance_rffs))
    for xdata, ldata in dataset.get_chunked_x_data():
        xfeatures = kernel.transform_x(xdata, ldata)
        z_trans_z += xfeatures[:,:variance_rffs].T @ xfeatures[:,:variance_rffs]
    return z_trans_z



"""Tests exact fitting (for small numbers of RFFs)."""
import sys
import os
import unittest

import numpy as np

sys.path.append(os.path.join(os.path.dirname(os.path.abspath(__file__)), ".."))
from utils.build_test_dataset import build_test_dataset
from utils.build_classification_dataset import build_discriminant_traintest_split
from utils.model_constructor import get_models
from utils.evaluate_model import evaluate_model

#A set of hyperparameters known to work well for our testing dataset
#that we can use as a default.
HPARAM = np.array([np.log(np.sqrt(0.0767)),  np.log(0.358)])
DISCRIM_HPARAM = np.array([0., -0.75])

NUM_RFFS = 2100


class CheckExactFit(unittest.TestCase):
    """Tests the exact fitting algorithm. We really just
    check to make sure there are no exceptions. For
    more extensive testing of performance using this
    and other fitting algorithms, see complete pipeline
    tests."""

    def test_exact_fit(self):
        """Test exact fitting."""
        online_data, _ = build_test_dataset(conv_kernel = False)
        cpu_mod, gpu_mod = get_models("RBF", online_data, num_rffs = NUM_RFFS)

        cpu_mod.fit(online_data, mode = "exact")

        if gpu_mod is not None:
            gpu_mod.fit(online_data, mode = "exact")


if __name__ == "__main__":
    unittest.main()



"""The noise hyperparameter (lambda) is actually the
ratio of two hyperparameters: alpha / beta. The choice
of alpha and beta does not however affect fitting or
variance as long as the ratio is fixed; it only
affects the marginal likelihood. This module
provides tools for optimizing the values of alpha
and beta for a given lambda so that the marginal
likelihood reported for a given lambda is as
favorable as possible."""
import numpy as np


def optimize_alpha_beta(lambda_, nll_terms, ndatapoints, nrffs,
        beta_max = 10., beta_min = 0.1):
    """Optimizes alpha and beta for a given value
    of lambda_ so that the marginal likelihood is
    at its best value for that setting for lambda_.

    Args:
        lambda_ (float): The ratio of alpha / beta.
        nll_terms (ndarray): A numpy array with two elements:
            -0.5 * y^T Zw + 0.5 * y^T y and 0.5 * ln|Z^T Z + lambda_ I|.
            These are used to calculate the marginal likelihood
            for different values of alpha and beta.
        ndatapoints (int): The number of datapoints.
        nrffs (int): The number of random features or features.
        beta_max (float): The max value for beta.
        beta_min (float): The min value for beta.

    Returns:
        score (float): The negative marginal log likelihood for
            the optimized alpha and beta.
        alpha_beta (ndarray): The optimized alpha and beta.
    """
    beta = np.sqrt(2 * nll_terms[0] / (ndatapoints * lambda_**2))
    beta = max(min(beta, beta_max), beta_min)
    score = nll_terms[0] / (beta * lambda_)**2 + (ndatapoints - nrffs) * np.log(lambda_)
    score += nll_terms[1] + ndatapoints * np.log(beta)
    return score + 0.5 * ndatapoints * np.log(2*np.pi), beta



"""This kernel is a fast Hadamard transform based convolutional
kernel that is the random features equivalent of the Ngram kernel."""
import numpy as np
from .conv_kernel_baseclass import ConvKernelBaseclass


class Conv1dRBF(ConvKernelBaseclass):
    """The Conv1d class is a convolutional kernel that can work
    with non-aligned time series or sequences.
    This class inherits from ConvKernelBaseclass.
    Only attributes unique to this child are described in this docstring.
    
    Attributes:
        hyperparams (np.ndarray): This kernel has two
            hyperparameters: lambda_ (noise)
            and sigma (inverse mismatch tolerance).
        bounds (np.ndarray): The builtin bounds for hyperparameter optimization,
            which can be overriden / reset by user.
    """

    def __init__(self, xdim, num_rffs, random_seed = 123, device = "cpu",
                    num_threads = 2, double_precision = False,
                    kernel_spec_parms = {}):
        """Constructor.

        Args:
            xdim (tuple): The dimensions of the input. Either (N, D) or (N, M, D)
                where N is the number of datapoints, D is number of features
                and M is number of timepoints or sequence elements (convolution
                kernels only).
            num_rffs (int): The user-requested number of random Fourier features.
                For sine-cosine kernels (RBF, Matern), this will be saved by the
                class as num_rffs.
            random_seed (int): The seed to the random number generator.
            device (str): One of 'cpu', 'gpu'. Indicates the starting device.
            num_threads (int): The number of threads to use for random feature generation
                if running on CPU. If running on GPU, this is ignored.
            double_precision (bool): If True, generate random features in double precision.
                Otherwise, generate as single precision.
            kernel_spec_parms (dict): A dictionary of additional kernel-specific
                attributes. In this case, should contain 'conv_width'.

        Raises:
            ValueError: A ValueError is raised if the dimensions of the input are
                inappropriate given the conv_width.
        """
        if "conv_width" not in kernel_spec_parms:
            raise ValueError("conv_width must be included as a kernel-specific "
                    "parameter if using a sequence kernel.")

        super().__init__(xdim, num_rffs, random_seed,
                num_threads, double_precision, kernel_spec_parms["conv_width"],
                kernel_spec_parms)
        self.hyperparams = np.ones((2))
        self.bounds = np.asarray([[1e-3,5], [1e-6, 1e2]])

        self.device = device



"""This module describes the OfflineDataset
for regression and classification.

The OfflineDataset bundles together attributes for handling
a dataset stored on disk. The dataset can return a specific
chunk of data from the list provided to it, can serve as a generator
and return all chunks in succession, or can provide a minibatch.
"""
import os
import copy

import numpy as np

from .data_handling_baseclass import DatasetBaseclass



class OfflineDataset(DatasetBaseclass):
    """The OfflineDataset class handles datasets which are stored on
    disk. Only unique attributes not shared by parent class are
    described here.

    Attributes:
        _xfiles (list): A list of absolute filepaths to the locations
            of each xfile. Each xfile is a numpy array saved as a .npy.
        _yfiles (list): A list of absolute filepaths to the locations
            of each yfile. Each yfile is a numpy array saved as a .npy.
        _sequence_lengths: Either None or a list of absolute filepaths
            to the locations of each .npy file containing the length
            of the sequence / graph for the corresponding datapoint.
    """
    def __init__(self, xfiles,
                       yfiles,
                       sequence_lengths,
                       xdim,
                       trainy_mean = 0.,
                       trainy_std = 1.,
                       max_class = 1,
                       chunk_size = 2000):
        """The class constructor for an OfflineDataset.

        Args:
            xfiles (list): A list of absolute filepaths to the locations
                of each xfile. Each xfile is a numpy array saved as a .npy.
            yfiles (list): A list of absolute filepaths to the locations
                of each yfile. Each yfile is a numpy array saved as a .npy.
            sequence_lengths: Either None or a list of absolute filepaths
                to the locations of each .npy file containing the length
                of the sequence / graph for the corresponding datapoint.
            xdim (list): A list of length 2 (for 2d arrays) or 3 (for 3d
                arrays) where element 0 is the number of datapoints in the
                dataset, element 1 is shape[1] and element 2 is shape[2]
                (for 3d arrays for convolution kernels only).
            trainy_mean (float): The mean of the y-values. Only used for
                regression.
            trainy_std (float): The standard deviation of the y-values.
                Only used for regression.
            max_class (int): The largest category number in the data. Only
                used for classification.
            device (str): The current device.
            chunk_size (int): The largest allowed file size (in # datapoints)
                for this dataset. Should be checked and enforced by caller.
        """
        super().__init__(xdim, chunk_size, trainy_mean,
                trainy_std, max_class)
        self._xfiles = [os.path.abspath(f) for f in xfiles]
        self._yfiles = [os.path.abspath(f) for f in yfiles]
        if sequence_lengths is not None:
            self._sequence_lengths = [os.path.abspath(f) for f in sequence_lengths]
        else:
            self._sequence_lengths = None


    def get_chunked_data(self):
        """A generator that returns the data stored in each
        file in the data list in order with paired x and y
        data."""
        if self._sequence_lengths is None:
            for xfile, yfile in zip(self._xfiles, self._yfiles):
                xchunk = np.load(xfile)
                ychunk = np.load(yfile).astype(np.float64)
                ychunk -= self._trainy_mean
                ychunk /= self._trainy_std
                yield xchunk, ychunk, None
        else:
            for xfile, yfile, lfile in zip(self._xfiles, self._yfiles,
                    self._sequence_lengths):
                xchunk, lchunk = np.load(xfile), np.load(lfile)
                ychunk = np.load(yfile).astype(np.float64)
                ychunk -= self._trainy_mean
                ychunk /= self._trainy_std
                yield xchunk, ychunk, lchunk



    def get_chunked_x_data(self):
        """A generator that returns the data stored in each
        file in the data list in order, retrieving x data
        and sequence length only."""
        if self._sequence_lengths is None:
            for xfile in self._xfiles:
                yield np.load(xfile), None
        else:
            for xfile, lfile in zip(self._xfiles, self._sequence_lengths):
                yield np.load(xfile), np.load(lfile)
